/**
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 *
 *   @author Robert Bakic (robert.bakic@gmail.com)
 */

package org.azuki.bean.exec ;

import java.io.OutputStream;
import java.util.*;

import org.azuki.bean.exec.server.imp.*;
import org.azuki.bean.exec.server.imp.mngr.ProcessManager;
import org.azuki.framework.annotation.*;
import org.azuki.framework.bean.Bean;
import org.azuki.framework.command.*;
import org.azuki.framework.command.ICommand.Priority;
import org.azuki.framework.context.Log;
import org.azuki.framework.injection.IMethodProxy;
import org.azuki.framework.injection.IMethodProxy.*;
import org.jdom.Element;

final public class ProcessFactoryBean
{
	final static private String outputName = "command";
		
	// EventPublisher injection
	@DependencyInjectionInfo(desc="Publish a message to an output connector",
							 param={Object.class, String.class, Object.class})
	public IMethodProxy<Void> publish ;

	@DependencyInjectionInfo(desc="Gets system env() value and user-defined variable for the external process execution")
	public IMethodProxy<Map<String, String>> getAllInherited ;
	
	// Safe-Accessor
	final public Map<String, String> getAllInherited()
	{
		if(getAllInherited==null) 
			return new HashMap<String, String>() ;
		
		try
		{
			return getAllInherited.call() ;
		}
		catch(Exception e)
		{
			return new HashMap<String, String>() ;
		}
	}
	
	private Execution execution ; // Singleton can be used, because thread-safe
	
	@MethodInfo(desc="Initializer contract implementation. Called at the framework initialization. Must not be called by others means")
    final public void postCreate(Element para)
	throws Exception
	{
		String key = Bean.getMetadata(this).getID().toString();
		PProcess.parseXmlToCommand(para, getAllInherited(), key);

		execution = PProcess.parseXmlToCommand(para, getAllInherited(), key);
	}
	
	@MethodInfo(desc="This method can receive events to create Jobs in the bean output")
	final public void processEvent()
	throws Exception
	{
		if(execution==null) return;
		
		publish.call(this, outputName, execution) ;
		Log.info("ProcessFactoryBean:processEvent::publish the execution to output ("+outputName+")");	
	}
	
	@Export
	@MethodInfo(desc="This method can be injected and call to create Jobs in the bean output. The command line is passed as argument. Note that the 'Command Parameter' is ingored in this case")
	final public void processCommand(String  command)
	throws Exception
	{
		if(execution==null) return;
		execution.setCmd(command);
		publish.call(this, outputName, execution) ;
		Log.info("ProcessFactoryBean:processCommand::publish the execution to output ("+outputName+")");	
	}
	
/////////////////////////////	
	final private ProcessManager pm = new ProcessManager(this) ;
	
	/*****************************
	 * Synchronous call : Invoke
	 *****************************/
	
	/**
	 * Executes the specified command.<br>
	 * You must wait for the return of the method.
	 * @param name The name of the command
	 * @param cmd  The command line array
	 * @param env  The new environment (if <i>null</i>, takes only the inherited environment)
	 * @param path An additional path to found the executable (look at first)
	 * @throws Exception
	 */
//A voir	@MethodExport
    final public void invoke(String name, String[] cmd, Map<String, String> env, String path)
	throws Exception
    {
    	pm.invoke(Command.mainQueue(), Priority.NORMAL, name, cmd, getAllInherited(), env, path, null) ;
    }

    /**
	 * Executes the specified command.<br>
	 * You must wait for the return of the method.
	 * @param name The name of the command
	 * @param cmd  The command line array
	 * @param env  The new environment (if <i>null</i>, takes only the inherited environment)
	 * @param path An additional path to found the executable (look at first)
	 * @param qname The name of the queue (dynamically created)
     * @param pty The priority (Priority.MAX, Priority.HIGH, Priority.NORMAL, Priority.LOW, Priority.MIN)
	 * @throws Exception
	 */
//  A voir	@MethodExport
    final public void invokeQueued(String name, String[] cmd, Map<String, String> env, String path, String qname, Priority pty)
	throws Exception
    {
    	pm.invoke(Command.getQueue(qname), pty, name, cmd, getAllInherited(), env, path, null) ;
    }

    /**
	 * Executes the specified command.<br>
	 * You must wait for the return of the method.
	 * @param name The name of the command (for the Chronometer)
	 * @param cmd  The command line array
	 * @param env  The new environment (if <i>null</i>, takes only the inherited environment)
	 * @param path An additional path to found the executable (look at first)
	 * @param out  An OutputStream for the output of the command (if <i>null</i>, the output is sent to stdout)
	 * @throws Exception
	 */
//  A voir	@MethodExport
	final public void invokeStreamed(String name, String[] cmd, Map<String, String> env, String path, OutputStream out)
	throws Exception
	{
		pm.invoke(Command.mainQueue(), Priority.NORMAL, name, cmd, getAllInherited(), env, path, out) ;
	}

    /**
	 * Executes the specified command.<br>
	 * You must wait for the return of the method.
	 * @param name The name of the command
	 * @param cmd  The command line array
	 * @param env  The new environment (if <i>null</i>, takes only the inherited environment)
	 * @param path An additional path to found the executable (look at first)
	 * @param qname The name of the queue (dynamically created)
     * @param pty The priority (Priority.MAX, Priority.HIGH, Priority.NORMAL, Priority.LOW, Priority.MIN)
	 * @param out  An OutputStream for the output of the command (if <i>null</i>, the output is sent to stdout)
	 * @throws Exception
	 */
//	A voir	@MethodExport
    final public void invokeQueuedStreamed(String name, String[] cmd, Map<String, String> env, String path, String qname, Priority pty, OutputStream out)
	throws Exception
    {
    	pm.invoke(Command.getQueue(qname), pty, name, cmd, getAllInherited(), env, path, out) ;
    }
    
	/*****************************
	 * ASynchronous call : Send
	 *****************************/
    
    /**
	 * Executes the specified command in another server thread.<br>
	 * You do not have to wait for the return of the method.
	 * @param name The name of the command 
	 * @param cmd  The command line array
	 * @param env The new environment (if <i>null</i>, takes the inherited environment)
	 * @param path An additional path to found the executable (look at first)
	 */
//  A voir	@MethodExport
	final public void send(String name, String[] cmd, Map<String, String> env, String path)
	{
		pm.send(Command.mainQueue(), Priority.NORMAL, name, cmd, getAllInherited(), env, path, null) ;
	}
    
    /**
     * Sends a command into a processing queue (with a priority for processing).
	 * @param name The name of the command (for the Chronometer)
	 * @param cmd The command array
	 * @param env The new environment (if <i>null</i>, takes the inherited environment)
	 * @param path The path to the command
	 * @param qname The name of the queue (dynamically created)
     * @param pty The priority (Priority.MAX, Priority.HIGH, Priority.NORMAL, Priority.LOW, Priority.MIN)
     */
//	A voir	@MethodExport
    final public void sendQueued(String name, String[] cmd, Map<String, String> env, String path, String qname, Priority pty)
    {
    	pm.send(Command.getQueue(qname), pty, name, cmd, getAllInherited(), env, path, null) ;
    }
    
    
    /**
     * Sends a command into a processing queue (with a priority for processing).
	 * @param name The name of the command (for the Chronometer)
	 * @param cmd The command array
	 * @param env The new environment (if <i>null</i>, takes the inherited environment)
	 * @param path The path to the command
     * @param cback A "callback" object
     */
//  A voir	@MethodExport
    final public void sendCallback(String name, String[] cmd, Map<String, String> env, String path, ICallback cback)
    {
    	pm.send(Command.mainQueue(), Priority.NORMAL, name, cmd, getAllInherited(), env, path, cback) ;
    }
    
    /**
     * Sends a command into a processing queue (with a priority for processing).
	 * @param name The name of the command (for the Chronometer)
	 * @param cmd The command array
	 * @param env The new environment (if <i>null</i>, takes the inherited environment)
	 * @param path The path to the command
	 * @param qname The name of the queue (dynamically created)
     * @param pty The priority (Priority.MAX, Priority.HIGH, Priority.NORMAL, Priority.LOW, Priority.MIN)
     * @param cback A "callback" object
     */
//  A voir	@MethodExport
    final public void sendQueuedCallback(String name, String[] cmd, Map<String, String> env, String path, String qname, Priority pty, ICallback cback)
    {
    	pm.send(Command.getQueue(qname), pty, name, cmd, getAllInherited(), env, path, cback) ;
    }		
}