/**
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 *
 *   @author Robert Bakic (robert.bakic@gmail.com)
 */

package org.azuki.bean.exec.server.imp ;

import java.io.File;
import java.io.OutputStream;
import java.util.*;

import org.azuki.bean.exec.server.imp.mngr.ProcessManager;
import org.azuki.bean.exec.shared.ConstProcessFactory;
import org.azuki.common.editor.AttrCommon;
import org.azuki.common.editor.TagCommon;
import org.azuki.common.util.JDOMUtil;
import org.azuki.framework.command.Command;
import org.azuki.framework.command.ICommand.Priority;
import org.azuki.framework.context.Log;
import org.jdom.Element;

final public class PProcess
{
	static volatile public int count = 0 ;
	
	
	// Rem : envp --> key to Environnement Parameters
    final static public void exec(ProcessManager pm, String name, String[] cmd, Map<String, String> env,
    		Map<String, String> nenv, String path, OutputStream out)
    throws Exception
    {
    	Integer ID = new Integer(++count) ;
        
        String cmdString = ExecUtil.toString(cmd) ;

        String nfo = new StringBuffer("-> START EXEC ID=").append(ID)
                     .append(" name=(").append(name)
                     .append(") cmd=(").append(cmdString)
                     .append(')').toString() ;

        Log.info(nfo);

        long t = System.currentTimeMillis() ; // Always because needed in PRunning
		
		try
		{
			// Process environnement
			String[] envp = calcEnv(env, nenv);
			
			// if path != null, we try to locate the programme using the path value.
			// then add the full location before the launching command.
			if(path!=null)
				lookAtPath(cmd, path) ;
			
			// If envp is null, the subprocess inherits the environment settings of the current process.
            // Actually cannot be null...
			Process proc = Runtime.getRuntime().exec(cmd, envp) ;
            
            // For console
			if(pm!=null)
				pm.addRunning(ID, name, cmd, proc, t);

            GTSync sync = new GTSync() ;

			// Push because this thread is already running
			Command.sendQueued(new GPReader(GPReader.OUT, proc.getInputStream(), out, sync), Command.mainQueue(), Priority.HIGHEST) ;
			Command.sendQueued(new GPReader(GPReader.ERR, proc.getErrorStream(), out, sync), Command.mainQueue(), Priority.HIGHEST) ;
			
			proc.waitFor();
			
			try
			{ proc.getOutputStream().close() ; }
			catch(Exception e)
			{}	
			
			int error = proc.exitValue();

            sync.waitFor(); // wait for output stream

			if(error != 0)
			{
				String mess = new StringBuffer("-- EXEC ID=")
								.append(ID).append(" cmd(")
								.append(cmdString).append(") RETURN ERROR : ")
								.append(error).toString() ;
								
				throw new Exception(mess) ;
			}
		}
		catch(Exception e)
		{
			// For supervisor
            if(pm!=null) 
            	pm.addError(ID, name, cmd, e.getMessage(), t, System.currentTimeMillis(), env) ;
			throw e ;	
		}
		finally
		{
			if(pm!=null)
				pm.removeRunning(ID);
			
			String mess = new StringBuffer("<- STOP  EXEC ID=").append(ID).toString() ; 
			Log.info(mess);
		}
	}
    
    // Final environnement parameter calculation
    final static private String[] calcEnv(Map<String, String> env, Map<String, String> nenv)
    throws Exception
    {
		// Add parameter environnment value.
		if(nenv!=null) ExecUtil.merge(env, nenv) ;
		
		String[] envp = new String[env.size()] ;
		
		int i=0 ;

		for(Map.Entry<String, String> ety : env.entrySet())
		{
			String key = ety.getKey() ;
			String val = ety.getValue() ;

			envp[i++] = new StringBuffer(key).append('=').append(val).toString() ;
		}
		
		return envp ;
    }

	static final private boolean win32 = System.getProperty("path.separator").equals(";") ;

	final static private void lookAtPath(String[] cmds, String path)
    {
    	for(int i=0; i<cmds.length; i++)	
    	{
			// Name of the batch
			String batch = cmds[i] ;
			
		    StringTokenizer st = new StringTokenizer(path, File.pathSeparator);
		    while(st.hasMoreTokens())
		    {
		    	String dir = st.nextToken().trim();
		        File df = new File(dir) ;
		        if(!df.exists()) continue ;
		        if(!df.isDirectory()) continue ;
		         
		        File f = new File(df, batch) ;
		        if(f.exists()) // Found
		        {
		        	cmds[i] = f.getAbsolutePath() ; // Change cmd with
		        	break ;
		        }
		        
		        if(!win32) continue ;
		        
		        // windoz, ugly...
		        f = new File(df, batch+".exe") ;
		        if(f.exists()) // Found
		        {  	cmds[i] = f.getAbsolutePath() ; // Change cmd with 
		        	break ; }

		        f = new File(df, batch+".bat") ;
		        if(f.exists()) // Found
		        {  	cmds[i] = f.getAbsolutePath() ; // Change cmd with 
		        	break ; }			    

		        f = new File(df, batch+".cmd") ;
		        if(f.exists()) // Found
		        {  	cmds[i] = f.getAbsolutePath() ; // Change cmd with 
		        	break ; }			    
		    }
		}    	
    }
	
	@SuppressWarnings("unchecked")
	final static public Execution parseXmlToCommand(Element para, Map<String,String> inherited, String beanKey)
	throws Exception
	{
		if(para==null) throw new Exception("Wrong parameters for process factory bean") ;
	
		Element exec = JDOMUtil.getElement(para, ConstProcessFactory.EXEC) ;
		if(exec==null) throw new Exception("Wrong parameters for process factory bean") ;
			
		String cmd = exec.getAttributeValue(ConstProcessFactory.COMMAND);
		if (cmd==null || "".equals(cmd)) throw new Exception("Wrong parameters for process factory bean") ;
			
		HashMap<String,String> nenv = new HashMap<String,String>() ;
		
		List<Element> lenv =exec.getChildren(TagCommon.ROW);
		for(Element lmen : lenv)
		{
			String name = lmen.getAttributeValue(AttrCommon.NAME) ;
			if(name==null) continue ;
			
			nenv.put(name, lmen.getAttributeValue(AttrCommon.VALU)) ;
		}
		
		String path = null ; // TODO : Add parameters
		OutputStream os = null ; // TODO : ?

		//String key = Bean.getMetadata(this).getID().toString();
		
		// PM !!!!!
		return  new Execution<Object>(null, beanKey, new String[] {cmd}, inherited, nenv, path, os) ;
	}
}