/**
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 *
 *   @author Mathieu Jucker
 */

package org.azuki.bean.jms;

import java.util.*;

import javax.jms.*;
import javax.jms.Queue;
import javax.naming.*;
import javax.servlet.http.*;

import org.azuki.bean.jms.server.console.JMSReaderConsole;
import org.azuki.bean.jms.shared.ConstJMS;
import org.azuki.common.editor.EditorUtil;
import org.azuki.framework.annotation.*;
import org.azuki.framework.bean.Bean;
import org.azuki.framework.context.Log;
import org.azuki.framework.injection.IMethodProxy;
import org.azuki.framework.injection.IMethodProxy.DependencyInjectionInfo;
import org.jdom.Element;

public class JMSReaderBean
{
	final static private String outputName = "Message";
	
	@DependencyInjectionInfo(desc="publish", param={Object.class,String.class,Object.class})
	public IMethodProxy publish;
	
	private String JNDI_CONTEXT_FACTORY;
	private String DESTINATION;
	
	private String CLIENT_ID;
	
	private String TIMEOUT;
	private int iTimeout;
	
	private boolean isValid = false;
	
	private ConnectionFactory qf;
	private Connection qc;
	private Session ses;
	private MessageConsumer mc;
	private Destination d;
	private Context JNDIContext;
	private TopicSubscriber tsus;	
	
	@MethodInfo(desc="Initializer contract implementation. Called at the framework initialization. Must not be called by others means")
	final public void postCreate(Element para)
	throws Exception
	{
		JNDIContext = new InitialContext();
		
		Map<String,String> map = EditorUtil.formToMap(para);
		if (map==null) throw new Exception("JMSReader::Bean parameters are missing !");
		
		DESTINATION=map.get(ConstJMS.KEY_DESTINATION);
		TIMEOUT=map.get(ConstJMS.KEY_TIMEOUT);
		JNDI_CONTEXT_FACTORY = map.get(ConstJMS.KEY_JNDI_CONTEXT_FACTORY);
		
		// Only mandatory for a topic
		CLIENT_ID=Bean.getMetadata(this).getID().toString() ;
		
		// check if parameters are valid
		if (JNDI_CONTEXT_FACTORY==null || JNDI_CONTEXT_FACTORY.isEmpty()) 
			throw new Exception("JMSReader::JNDI_CONTEXT_FACTORY name is not set : can't get any ContextFactory");
		
		
		if (DESTINATION==null || DESTINATION.isEmpty()) 
			throw new Exception("JMSReader::QUEUE name is not set : can't read any queue");
		
		
		if (TIMEOUT==null || TIMEOUT.isEmpty())
		{
			Log.warning("JMSReader::timeout is not set : using 100ms as default value");
			TIMEOUT="100";
		}
		
		try
		{
			iTimeout = Integer.parseInt(TIMEOUT);
		}
		catch(NumberFormatException ne)
		{
			Log.warning("JMSReader::timeout parameter does not contain an integer value : using 100ms as default value");
			iTimeout=100;
		}
		
		isValid = true;
		
	}
	
	@Export
	public Object getMessageFromDesination() 
	throws Exception
	{
		
		try
		{
			if (!isValid) 
			{
				closeAll();
				throw new Exception("JMSReader::parameters are invalid");
			}

			qf = (ConnectionFactory) JNDIContext.lookup("java:/comp/env/"+JNDI_CONTEXT_FACTORY);
			if (qf==null ) throw new Exception("JMSReader::JMS ContextFactory is null : can't get a JMS Connection");
			
			d = (Destination) JNDIContext.lookup("java:/comp/env/"+DESTINATION);
			if (d==null)  throw new Exception("JMSReader::JMS Destination is null : JNDI lookup failed");
						
			initConnection();
		
			Message tm=null;
			if (tsus!=null)
			{
				tm = tsus.receive(iTimeout);
				ses.commit();
			}
			else
			{
				tm = mc.receive(iTimeout);
				ses.commit();
			    mc.close();
			}
			Object msg = null;
			

			if (tm==null )
			{
				Log.info("no message in destination");
			}
			else
			{
				if (tm instanceof ObjectMessage)
				{
					ObjectMessage om = (ObjectMessage) tm;
					msg = om.getObject();
				}
				
				else msg = tm;
			}
			
			closeAll();
			return msg;
		}
		catch (Exception e)
		{
			Log.severe("JMSReader::error occured while trying to read a message from broker") ;
			throw e;
		}
	}

	
	final private void initConnection() throws Exception
	{
		if (qf==null ) throw new Exception("JMSReader::JMS ContextFactory is null : can't get a JMS Connection");
		
		qc = qf.createConnection();
		if (qc==null)	throw new Exception("JMSReader::JMS Connection is null : createConnection failed");
		
		qc.setClientID(CLIENT_ID);

		if (d==null) throw new Exception("JMSReader::JMS Destination is null : JNDI lookup failed");
		
		if (d instanceof Topic)
		{
			Topic t = (Topic) d;
			ses = qc.createSession(true,Session.AUTO_ACKNOWLEDGE);
			tsus = ses.createDurableSubscriber(t,CLIENT_ID);
		}
		else
		{
			ses = qc.createSession(true,Session.AUTO_ACKNOWLEDGE);
		}

		if (ses==null)
		{
			Log.severe("JMSReader::JMS Session is null : createSession failed");
			closeAll();
			throw new Exception("JMSReader::JMS Session is null : createSession failed");
		}
		
		qc.start();
		mc = ses.createConsumer(d);
		
	}
	
	final private void closeAll() throws JMSException
	{
		if (tsus!=null)
			tsus.close();
		if (ses!=null)
			ses.close();
		if (qc!=null)
		{
			qc.stop();
			qc.close();
		}
	}
	
	@MethodInfo(desc="The contract implementation of the application portal aspect (AppPortal) for the Console application")
	final public void console(HttpServlet srv, HttpServletRequest req, HttpServletResponse res)
	throws Exception
	{
		JMSReaderConsole.service(this,srv, req, res) ;
	}

	@Export
	final public Vector<Message> dumpQueue()
	throws Exception
	{
		if (!(d instanceof Queue))
			throw new Exception("JMSReader::destination only could be a Queue");
		
		initConnection();
		
		Queue queue = null;
		QueueBrowser qb = null;
		
		queue = (Queue) d;
		qb = ses.createBrowser(queue);
		Enumeration enu = qb.getEnumeration();
		
		Vector<Message> v = new Vector<Message>();
		if (enu!=null && enu.hasMoreElements())
		{
			while (enu.hasMoreElements())
			{
				Message msg = (Message) enu.nextElement();
				v.add(msg);
			}
		}
		
		qb.close();
		closeAll();
	    return v;
	}
	
	// get a message from destination and publish it into an output
	public void getMsgAndPublishEvent()
	throws Exception
	{
		Object msg = this.getMessageFromDesination();
		
		// don't publish a null object
		if (msg==null) return;
		publish.call(this,outputName,msg);
	}
}
