/**
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 *
 *   @author Mathieu Jucker
 *   @author Yoann Perrin (yoann.perrin@gmail.com)
 */

package org.azuki.bean.jmx.server;

import java.lang.reflect.Field;

import javax.management.*;

import org.azuki.bean.jmx.server.imp.*;
import org.azuki.bean.jmx.shared.ConstJMX;
import org.azuki.framework.aop.IPointcut;
import org.azuki.framework.bean.IMetadata;
import org.azuki.framework.context.*;
import org.azuki.framework.injection.*;
import org.azuki.framework.object.Locator;

public class JMXExportAdvice implements IController, DynamicMBean, NotificationEmitter
{
	private ObjectName onam ;
	private MBeanInfo mBeanInfo ;  // Lazy
	
	private NotificationBroadcasterSupport notfCaster = new NotificationBroadcasterSupport() ;
	private int notfSeq = 0 ;
	
	final private IPointcut ipc ;
	
	// constructor
	public JMXExportAdvice(IPointcut ipc, String domain)
	throws Exception
	{
		this.ipc = ipc ;
		
		IMetadata meta = ipc.getTargetMetadata() ;
		
		String labl = meta.getLabel() ;
		if(labl==null) labl="";
		
		onam = ObjectNameFactory.create(domain, meta.getName(), meta.getLabel(), meta.getID()) ;
	}

	final public ObjectName getObjectName()
	{
		return onam ;
	}
	

	/*
	 * Events
	 */
	final protected void fireAttributeChange(String name, String type, Object oldValue, Object newValue)
	{
		long time = System.currentTimeMillis() ; 
		String mess = "Attribute change: "+name+" old value("+oldValue+") new value("+newValue+")" ;
		
		AttributeChangeNotification notf = new AttributeChangeNotification(this, ++notfSeq, time, mess, name, type, oldValue, newValue) ;
		
		notify(notf) ;
	}
	
	final protected void fireMethodCalled(String name, Object source)
	{
		MethodCallNotification notf = new MethodCallNotification(source, ++notfSeq, this, name) ;
		
		notify(notf) ;
	}
	
	final private void notify(Notification notf)
	{
		notfCaster.sendNotification(notf) ;
	}
	
	/*
	 * DynamicMBean
	 */
	final public Object getAttribute(String attr)
	throws AttributeNotFoundException
	{
		try
		{
			return JMXReflect.getFieldValue(this, attr) ;
		}
		catch(Exception ex)
		{
			Log.severe(Log.getStack(ex)) ;
			throw new AttributeNotFoundException(ex.getMessage()) ;
		}
	}

	final public void setAttribute(Attribute attribute)
	throws AttributeNotFoundException
	{
		String attrName = attribute.getName() ;
		Object attrValu = attribute.getValue() ;
		
		try
		{
			Field field = JMXReflect.getField(this, attrName) ;
			Object oldValue = JMXReflect.getFieldValue(this, field) ;
			
			JMXReflect.setFieldValue(this, field, attrValu) ;
			
			fireAttributeChange(attrName, field.getType().getName(), oldValue, attrValu) ;
		}
		catch(Exception ex)
		{
			Log.severe(Log.getStack(ex)) ;
			throw new AttributeNotFoundException(ex.getMessage()) ;
		}
	}

	final public AttributeList getAttributes(String[] attrNames)
	{
		AttributeList list = new AttributeList() ;
		if(attrNames.length==0) return list ;
		
		for(String name : attrNames)
		{
			Object valu = null ;
			try
			{ valu = getAttribute(name) ; }
			catch(Exception ex)
			{ /*Log.throwing(getClass().getName(), "getAttributes", ex) ;*/ }
			
			list.add(new Attribute(name, valu)) ;
		}
		
		return list;
	}

	final public AttributeList setAttributes(AttributeList attributes)
	{
		AttributeList list = new AttributeList() ;
		if(attributes.isEmpty()) return list ;
		
		for(Attribute attr : attributes.asList())
		{
			try
			{
				setAttribute(attr) ;
				list.add(attr) ;
			}
			catch(Exception ex)
			{
				/*Log.throwing(getClass().getName(), "getAttributes", ex) ;*/
			}
		}
		
		return list;
	}

	// invoke by JMX
	final public Object invoke(String actionName, Object[] params, String[] signature)
	throws ReflectionException
	{
		try
		{
			if(!JMXUtil.contains(ipc.getSelectedMethods(), actionName)) return null ;
	
			try
			{
				IMetadata meta = ipc.getAspectMetadata() ;
				
				IMetadata tmeta = ipc.getTargetMetadata() ;
				Object[] args = {tmeta.getName(), tmeta.getLabel(), actionName, params} ;
				
				Thread.currentThread().setContextClassLoader(JMXExportAdvice.class.getClassLoader()) ;
				
				return ((IMethodProxy)Locator.create(Locator.OBJECT_METHOD_PROXY, meta.getName(), meta.getLabel(), ConstJMX.INVOKE)).call(args) ; 
			}
			finally
			{
				ThreadContext.reset() ; // Clean thread context
			}
			// Notification with --> IController.intercept()
			//fireMethodCalled(actionName, this) ;
		}
		catch(Exception e)
		{
			e.printStackTrace();
			Log.severe(Log.getStack(e)) ;
			throw new ReflectionException(e);			
		}
		catch(Throwable th)
		{
			th.printStackTrace();
			Log.severe(Log.getStack(th)) ;
			throw new ReflectionException(new Exception(Log.getStack(th)));
		}
	}

	final public MBeanInfo getMBeanInfo()
	{
		if(mBeanInfo!=null) return mBeanInfo ; 
		
		mBeanInfo = MBeanInfoFactory.mkInfo(ipc) ; // lazy

		return mBeanInfo ;

	}
	
	final public void removeNotificationListener(NotificationListener listener, NotificationFilter filter, Object handback)
	throws ListenerNotFoundException
	{
		notfCaster.removeNotificationListener(listener, filter, handback) ;
	}

	final public void addNotificationListener(NotificationListener listener, NotificationFilter filter, Object handback)
	throws IllegalArgumentException
	{
		notfCaster.addNotificationListener(listener, filter, handback) ;
	}

	final public void removeNotificationListener(NotificationListener listener)
	throws ListenerNotFoundException
	{
		notfCaster.removeNotificationListener(listener) ;
	}

	final public MBeanNotificationInfo[] getNotificationInfo()
	{
		return getMBeanInfo().getNotifications() ;
	}

	/*
	 * IController
	 */
	final public Object intercept(InvocationTarget it) 
	throws Exception
	{
		Object retour = it.proceed() ;
		String method = it.getMethodName() ;
		
		if(JMXUtil.contains(ipc.getSelectedMethods(), method))
			fireMethodCalled(method, this) ;
		
		return retour;
	}
}
