/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.bean.ooo ;

import java.io.*;
import java.util.*;

import net.sf.jooreports.templates.*;

import org.azuki.bean.ooo.server.imp.OOImageReplacer;
import org.azuki.bean.ooo.shared.*;
import org.azuki.common.util.*;
import org.azuki.framework.annotation.*;
import org.azuki.framework.context.Log;
import org.azuki.framework.injection.IMethodProxy;
import org.azuki.framework.injection.IMethodProxy.*;
import org.jdom.Element;
import org.xml.sax.InputSource;

import freemarker.ext.dom.NodeModel;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */
public class OOReportingBean
{
	final static private String OUTPUT_NAME = "out" ;
	
	@DependencyInjectionInfo(desc="To convert document", param={String.class, byte[].class, String.class})
	public IMethodProxy<byte[]> convert ;
	
	@DependencyInjectionInfo(desc="The method used to publish a message to an output connector",
							param={Object.class, String.class, Object.class})
	public IMethodProxy<Void> publish ;
	
	private boolean publ = false ;
	final public boolean isPublished() 				{ return publ ; }
	
	private String tmplName = null ;
	final public String getTemplateName() 			{ return tmplName ; }
	
	private DocumentTemplate tmpl = null ; 
	final public DocumentTemplate getTemplate() 	{ return tmpl ; }
	
	private Map<String,byte[]> imap = new HashMap<String,byte[]>() ;
	final public Map<String,byte[]> getImageMap()	{ return imap ; }
	
	@MethodInfo(desc="Initializer contract implementation. Called at the framework initialization. Must not be called by others means")
	final public void postCreate(Element lmen)
	throws Exception
	{
		Element repo = JDOMUtil.getElement(lmen, ConstOOO.REPO) ;
		
		// template name
		tmplName = repo.getAttributeValue(ConstOOO.REPO_TMPL) ;
		if(!StringUtil.isEmpty(tmplName))
		{
			try
			{
				InputStream is = UtilOOO.getResourceAsStream(tmplName) ;
				if(is==null) throw new IOException("Not found: "+tmplName) ;
				
				//TODO: adapter le template au type de document !!! 
				tmpl = new ZippedDocumentTemplate(is) ;
			}
			catch(Exception ex)
			{
				Log.severe("Failed to load document template ("+tmplName+"): "+ex.getMessage()) ;
			}
			finally
			{
				if(tmpl==null) throw new Exception("Template ("+tmplName+") not found") ;
			}
		}
		
		// publish result ?
		publ = Util.isOK(repo.getAttributeValue(ConstOOO.REPO_PUBL)) ;
		
		// images
		List<Element> imags = JDOMUtil.listChildren(repo, ConstOOO.REPO_IMAG) ;
		for(Element imag : imags)
		{
			String name = imag.getAttributeValue(ConstOOO.REPO_IMAG_NAME) ;
			String valu = imag.getAttributeValue(ConstOOO.REPO_IMAG_VALU) ;
			
			if(StringUtil.isEmpty(valu))
			{	imap.put(name, UtilOOO.emptyImage()) ;
				continue ; }
			
			byte[] byts = null ;
			
			try
			{ byts = UtilOOO.getResource(valu) ; }
			catch(Exception ex)
			{ Log.severe("Image not found: "+valu) ; }
			finally
			{ if(byts!=null) imap.put(name, byts) ; }
		}
	}
		
	@Export
	@MethodInfo(desc="Generate report document from configured template with specified data",
				retu="The generated document bytes")
	final public byte[] report(@ArgInfo(name="byts", desc="Data document bytes")byte[] byts,
							   @ArgInfo(name="outf", desc="Generated document file name")String outf)
	throws Exception
	{
		return reportWithMapping(byts, outf, null) ;
	}
	
	@Export
	@MethodInfo(desc="Generate report document from configured template with specified data and image mapping",
				retu="The generated document bytes")
	final public byte[] reportWithMapping(@ArgInfo(name="byts", desc="Data document bytes")byte[] byts,
										  @ArgInfo(name="outf", desc="Generated document file name")String outf,
										  @ArgInfo(name="mapg", desc="Images mapping (key, resource name)")Map<String,String> mapg)
	throws Exception
	{
		if(mapg==null)
			return reportWithMappingBytes(byts, outf, null) ;
		
		Map<String,byte[]> mapb = new HashMap<String,byte[]>() ;
		
		for(Map.Entry<String,String> ety: mapg.entrySet())
		{
			String name = ety.getKey() ;
			String valu = ety.getValue() ;
			
			if(StringUtil.isEmpty(valu))
			{	mapb.put(name, UtilOOO.emptyImage()) ;
				continue ; }
			
			byte[] imgb = null ;
			
			try
			{ imgb = UtilOOO.getResource(valu) ; }
			catch(Exception ex)
			{ Log.severe("Image not found: "+valu) ; }
			finally
			{ if(byts!=null) mapb.put(name, imgb) ; }
		}
		
		return reportWithMappingBytes(byts, outf, mapb) ;
	}
	
	@Export
	@MethodInfo(desc="Generate report document from configured template with specified data and image mapping",
				retu="The generated document bytes")
	final public byte[] reportWithMappingBytes(@ArgInfo(name="byts", desc="Data document bytes")byte[] byts,
										  	   @ArgInfo(name="outf", desc="Generated document file name")String outf,
										  	   @ArgInfo(name="mapg", desc="Images mapping (key, resource bytes)")Map<String,byte[]> mapg)
	throws Exception
	{
		return generateImpl(tmpl, byts, outf, mapg) ;
	}
	
	@Export
	@MethodInfo(desc="Generate report document from specified template and data",
				retu="The generated document bytes")
	final public byte[] generate(@ArgInfo(name="tmpl", desc="Template document bytes")byte[] tmpl,
								 @ArgInfo(name="data", desc="Data document bytes")byte[] data,
								 @ArgInfo(name="outf", desc="Generated document file name")String outf)
	throws Exception
	{
		return generateWithMapping(tmpl, data, outf, null) ;
	}
	
	@Export
	@MethodInfo(desc="Generate report document with specified template, data and image mapping",
				retu="The generated document bytes")
	final public byte[] generateWithMapping(@ArgInfo(name="tmpl", desc="Data document bytes")byte[] tmpl,
											@ArgInfo(name="data", desc="Data document bytes")byte[] data,
								 			@ArgInfo(name="outf", desc="Generated document file name")String outf,
								 			@ArgInfo(name="mapg", desc="Images mapping (key, resource name)")Map<String,String> mapg)
	throws Exception
	{
		if(mapg==null)
			return generateWithMappingBytes(tmpl, data, outf, null) ;
		
		Map<String,byte[]> mapb = new HashMap<String,byte[]>() ;
		
		for(Map.Entry<String,String> ety: mapg.entrySet())
		{
			String name = ety.getKey() ;
			String valu = ety.getValue() ;
			
			if(StringUtil.isEmpty(valu))
			{	mapb.put(name, UtilOOO.emptyImage()) ;
				continue ; }
			
			byte[] imgb = null ;
			
			try
			{ imgb = UtilOOO.getResource(valu) ; }
			catch(Exception ex)
			{ Log.severe("Image not found: "+valu) ; }
			finally
			{ if(imgb!=null) mapb.put(name, imgb) ; }
		}
		
		return generateWithMappingBytes(tmpl, data, outf, mapb) ;
	}
	
	@Export
	@MethodInfo(desc="Generate report document with specified template, data and image mapping",
				retu="The generated document bytes")
	final public byte[] generateWithMappingBytes(@ArgInfo(name="tmpl", desc="Data document bytes")byte[] tmpl,
												 @ArgInfo(name="data", desc="Data document bytes")byte[] data,
												 @ArgInfo(name="outf", desc="Generated document file name")String outf,
												 @ArgInfo(name="mapg", desc="Images mapping (key, resource bytes)")Map<String,byte[]> mapg)
	throws Exception
	{
		ByteArrayInputStream bais = new ByteArrayInputStream(tmpl) ;
		
		try
		{ return generateImpl(new ZippedDocumentTemplate(bais), data, outf, mapg) ; }
		finally
		{ if(bais!=null) bais.close() ; }
	}
	
	final private byte[] generateImpl(DocumentTemplate tmpl, byte[] byts, String outf, Map<String,byte[]> imgs)
	throws Exception
	{
		Log.info("Generating document("+outf+") using template("+tmplName+")") ;
		
		// building image mapping
		Map<String,byte[]> imap = new HashMap<String,byte[]>(this.imap) ;
		if(imgs!=null) imap.putAll(imgs) ;
		
		// parsing extension
		String oute = outf ;
		int idx = oute.lastIndexOf('.') ;
		if(idx!=-1) oute = oute.substring(idx+1) ;
		
		// data
		ByteArrayInputStream bais = new ByteArrayInputStream(byts) ;
		ByteArrayOutputStream baos = new ByteArrayOutputStream() ;
		
		NodeModel data = NodeModel.parse(new InputSource(bais)) ; // only XML data...
		
		// generate document
		tmpl.createDocument(data, baos) ; // DON'T USE IMAGE PROVIDER !!
		
		try { bais.close() ; baos.close() ; }
		catch(Exception ex){}
		
		byte[] outb = baos.toByteArray() ;
		outb = OOImageReplacer.replace(outb, imap) ;
		
		// convert document if needed
		boolean cvrt = oute!=null && !oute.toLowerCase().equals(ConstOOO.ODT_EXTENSION) ;
		
		if(cvrt && convert!=null)
			outb = convert.call(ConstOOO.ODT_EXTENSION, outb, oute) ;
		
		// publish document if needed
		//TODO: post document into XML structure !
		if(publ && publish!=null)
		{
			Log.info("Publish document("+outf+") on output("+OUTPUT_NAME+")") ;
			
			try					{ publish.call(this, OUTPUT_NAME, outb) ; }
			catch(Exception ex)	{ Log.severe(Log.getStack(ex)) ; }
		}
		
		return outb ;
	}
}