/**
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 *
 *   @author Robert Bakic (robert.bakic@gmail.com)
 */

package org.azuki.bean.request ;

import java.lang.reflect.Method;
import java.util.*;

import javax.servlet.http.*;

import org.azuki.bean.request.server.RequestHandlerAdvice;
import org.azuki.bean.request.server.console.RequestHandlerConsole;
import org.azuki.bean.request.server.imp.*;
import org.azuki.bean.request.shared.ConstRequestHandler;
import org.azuki.common.bean.rpc.Service;
import org.azuki.common.util.Util;
import org.azuki.framework.annotation.*;
import org.azuki.framework.aop.*;
import org.azuki.framework.bean.*;
import org.azuki.framework.context.Log;
import org.jdom.Element;

final public class RequestHandlerBean implements IAspect<RequestHandlerAdvice>
{
	final static public String NAME = "RequestDispatcher" ;
	final static public String LABL = "main" ;
	
	static private String servletContextPath = "" ;
	final static public void setContextPath(String path){ servletContextPath = path ; }
	
	private IMetadata meta = null ; // cached
	final public IMetadata getMetadata()
	throws Exception
	{
		if(meta==null) meta = Bean.getMetadata(this) ;
		return meta ;
	}
	
	// Store here 
	private IPointcut ipcRD = null ; // can be null
	final public IPointcut getParentPC()
	{
		try
		{
			if(ipcRD!=null) return ipcRD ;
			
			IMetadata meta = Bean.getMetadata(this) ;
			List<IPointcut> lpc = meta.getPointcuts(NAME) ;
			// Assuming pointcut applied only once.
			if(lpc.size()>0) ipcRD = lpc.get(0);
			
			return ipcRD ;
		}
		catch(Exception e) { return null ; }
	}
	
	@MethodInfo(desc="Called at the framework initialisation to apply a pointcut. Must not be called by others means")
	final public RequestHandlerAdvice createAdvice(IPointcut ipc)
	throws Exception
	{
		return RequestHandlerAdvice.create(this, ipc) ;
	}

	@MethodInfo(desc="The contract implementation of the application portal aspect (AppPortal) for the Console application")
	final public void console(HttpServlet srv, HttpServletRequest req, HttpServletResponse res)
	throws Exception
	{
		RequestHandlerConsole.service(this, srv, req, res) ;
	}	

	// Called by the bridge servlet or other dispatcher
	@MethodInfo(desc="The contract implementation of the request dispatcher aspect (RequestHandlerBean)")
    final public void service(HttpServlet srv, HttpServletRequest req, HttpServletResponse res)
    throws Exception
    {
    	PDispatch.process(this, srv, req, res) ;
    }	

	// Service Part
	////////////////////////////////////////////////////////
	// All objects below respect the RequestHandler contract
    // Strings are always full URIs
	final private Map<String, Service> exact = new HashMap<String, Service>() ;
	final private Map<String, Service> start = new HashMap<String, Service>() ;
	final private Map<String, Service> end   = new HashMap<String, Service>() ;
	final private Map<String, Service> match = new HashMap<String, Service>() ;	
	
	
	@Export
	@MethodInfo(desc="The class name of the Servlet defined as a bridge between Azuki Framework and the servlet container",
				retu="The name of the bridge servlet")
	final public String getBridgeServletName()
	{
		return BridgeServlet.class.getName() ;
	}
	
	@Export
	@MethodInfo(desc="Retrieves a map of all the bean methods exactly bind to URIs ",
			retu="A map with an URI as key and Service object as value")
	final public Map<String, Service> getExactBindings()
	{ 
		return exact; 
	}
	
	@Export
	@MethodInfo(desc="Retrieves a map of all the bean methods bind to URLs starting with the specified URI",
			retu="A map with an URI as key and Service object as value")
	final public Map<String, Service> getStartBindings()
	{ 
		return start; 
	}
	
	@Export
	@MethodInfo(desc="Retrieves a map of all the bean methods bind to URLs ending with the specified text",
			retu="A map with the text as key and Service object as value")
	final public Map<String, Service> getEndBindings()
	{ 
		return end; 
	}
	
	@Export
	@MethodInfo(desc="Retrieves a map of all the bean methods bind to URLs containing the specified text",
			retu="A map with the text as key and Service object as value")
	final public Map<String, Service> getMatchBindings()
	{ 
		return match; 
	}
	
	@Export
	@MethodInfo(desc="Binds a bean method with the specified relative URI",
			    retu="The full URI bound with the bean method.")
	final public String bindExact(
			@ArgInfo(name="uri", desc="The URI. This URI is relative to the dispatcher's one") String uri, 
			@ArgInfo(name="bid", desc="The Bean'ID containing the specified method") Integer bid, 
			@ArgInfo(name="method", desc="The bean's method that will be bound") Method method)
	throws Exception
	{
		Log.info("BindEXACT : " + uri ) ;
		
		String fulluri = new StringBuilder(getBaseURI()).append(slashed(uri)).toString() ;
		if(exact.containsKey(fulluri))
			throw new Exception("Already binded: "+fulluri) ;
		
		IMetadata meta = Bean.getMetadata(bid) ;		
		exact.put(fulluri, new Service(fulluri, meta.getName(), meta.getLabel(), method)) ;
		return fulluri ;
	}

	@Export
	@MethodInfo(desc="Binds a bean method with all the URL starting with a specified URI",
		    retu="The full URI bound with the bean method.")
	final public String bindStart(
			@ArgInfo(name="uri", desc="The URI. This URI is relative to the dispatcher's one") String uri, 
			@ArgInfo(name="bid", desc="The Bean'ID containing the specified method") Integer bid, 
			@ArgInfo(name="method", desc="The bean's method that will be bound") Method method)
	throws Exception
	{
		Log.info("BindSTART : " + uri ) ;
		
		String fulluri = new StringBuilder(getBaseURI()).append(slashed(uri)).toString() ;
		if(start.containsKey(fulluri))
			throw new Exception("Already binded: "+fulluri) ;
		
		IMetadata meta = Bean.getMetadata(bid) ;
		start.put(fulluri, new Service(fulluri, meta.getName(), meta.getLabel(), method)) ;
		return fulluri ;
	}	
	
	@Export
	@MethodInfo(desc="Binds a bean method with all the URL ending with the specified suffix")
	final public void bindEnd(
	@ArgInfo(name="suffix", desc="The URL's suffix bound with the bean method") String suffix, 
	@ArgInfo(name="bid", desc="The Bean'ID containing the specified method") Integer bid, 
	@ArgInfo(name="method", desc="The bean's method that will be bound") Method method)
	throws Exception
	{
		Log.info("BindEND : " + suffix ) ;
		if(end.containsKey(suffix))
			throw new Exception("Already binded: "+suffix) ;
		
		IMetadata meta = Bean.getMetadata(bid) ;
		end.put(suffix, new Service(suffix, meta.getName(), meta.getLabel(), method)) ;
	}	

	@Export
	@MethodInfo(desc="Binds a bean method with all the URL containing the specified text")
	final public void bindMatch(
			@ArgInfo(name="text", desc="The text that must be contained in the URL") String text, 
			@ArgInfo(name="bid", desc="The Bean'ID containing the specified method") Integer bid, 
			@ArgInfo(name="method", desc="The bean's method that will be bound") Method method)
	throws Exception
	{
		Log.info("BindMATCH : " + text ) ;
		if(match.containsKey(text))
			throw new Exception("Already binded: "+text) ;
		
		IMetadata meta = Bean.getMetadata(bid) ;
		match.put(text, new Service(text, meta.getName(), meta.getLabel(), method)) ;
	}	
	
	
	@Export
	@MethodInfo(desc="Returns all URI bound to a bean method",
				retu="All the URI bound to the bean method. Note : returns parts of URI depending the way the method is bound ('startwith', 'endwith' and 'match'" )
	final public String[] getBindingURI(
	@ArgInfo(name="bid", desc="The Bean'ID containing the specified method") Integer bid, 
	@ArgInfo(name="method", desc="The name of the bean's method bound") String meth)
	throws Exception
	{
		Log.finest(" BID : " + bid ) ;
		Log.finest(" MET : " + meth ) ;
		
		IMetadata meta = Bean.getMetadata(bid) ;
		
		String name = meta.getName() ;
		String labl = meta.getLabel() ;
		
		List<String> als = new ArrayList<String>() ;
		
		addList(als, name, labl, meth, exact.entrySet()) ;
		addList(als, name, labl, meth, start.entrySet()) ;
		addList(als, name, labl, meth, end.entrySet()) ;
		addList(als, name, labl, meth, match.entrySet()) ;
		
		Log.finest(" VAL SIZE : " + als.size() ) ;
		
		return als.toArray(new String[als.size()]) ;
	}
	
	final static private void addList(List<String> als, String name, String labl, String meth, Collection<Map.Entry<String, Service>> col)
	{
		for(Map.Entry<String, Service> val : col)
		{	Service srv = val.getValue() ;
			Log.finest("x ("+Util.oneString(srv.name, srv.labl)+") y ("+srv.meth.getName()+")") ;
			if(isSameMethod(name, labl, meth, srv)) als.add(val.getKey()) ; }		
	}
	
	final static private boolean isSameMethod(String name, String labl, String meth, Service srv)
	{
		if(!name.equals(srv.name)) return false ;
		if(!labl.equals(srv.labl)) return false ;
		if(!meth.equals(srv.meth.getName())) return false ;
		return true ;		
	}
	
	// uri must startWith "/"
	final public Service get(String uri)
	{
		Service service = exact.get(uri) ;
		if(service!=null) return service ;
		
		for(Map.Entry<String,Service> strt : start.entrySet())
		{
		    if(uri.startsWith(strt.getKey()))
		        return strt.getValue() ;
		}
		
		for(Map.Entry<String,Service> strt : end.entrySet())
		{
		    if(uri.endsWith(strt.getKey()))
		        return strt.getValue() ;
		}

		for(Map.Entry<String,Service> strt : match.entrySet())
		{
		    if(uri.contains(strt.getKey()))
		        return strt.getValue() ;
		}

		return null ; // Default handler
	}	
	
	// Convert an URI "/supervisor.gml" or "supervisor.gml"
	// into an URI prefix by /: "/supervisor.gml". 
	final public String slashed(String uri)
	{
	    if(uri.startsWith("/")) return uri ;
	    return new StringBuilder("/").append(uri).toString() ;
	}	
	
	// Tip : based uri has to be found in code advice parameter....
	private String baseuri=null ;
	final private String getBaseURI()
	{
		if(baseuri!=null) return baseuri ;

		IPointcut ppc = getParentPC() ;
		
		if(ppc==null) 
		{	baseuri = servletContextPath ; // Main
			return baseuri ; }
		
		Element lm = ppc.getParameter() ;
		if(lm==null)
		{	baseuri = servletContextPath ;
			return baseuri ; }
		
		lm = lm.getChild(ConstRequestHandler.URI) ;
		if(lm==null) 
		{	baseuri = servletContextPath ;
			return baseuri ; }
		
		String uri = slashed(lm.getAttributeValue(ConstRequestHandler.NAME, "")) ;		
		
		baseuri = getParentURI()+uri ;
		
		return baseuri ;
	}
	
	final private String getParentURI()
	{
		IPointcut ppc = getParentPC() ;
		
		if(ppc==null) return "" ;
		
		try
		{
			RequestHandlerBean rq = (RequestHandlerBean)ppc.getAspectMetadata().getBean() ;
			return rq.getBaseURI() ;
		}
		catch(Exception e)
		{
			Log.severe(Log.getStack(e)) ;
			return "" ;
		}
	}
}