/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.bean.tomcat ;

import java.io.*;
import java.net.BindException;
import java.util.*;

import javax.servlet.http.*;

import org.apache.catalina.*;
import org.apache.catalina.connector.Connector;
import org.apache.catalina.startup.Embedded;
import org.azuki.bean.tomcat.server.console.TomcatEngineConsole;
import org.azuki.bean.tomcat.server.imp.*;
import org.azuki.bean.tomcat.shared.ConstTomcat;
import org.azuki.common.util.JDOMUtil;
import org.azuki.framework.Framework;
import org.azuki.framework.annotation.*;
import org.azuki.framework.context.Log;
import org.jdom.Element;

/**
 *   @author Yoann Perrin
 */
final public class TomcatEngineBean
{
	final private List<TomcatServlet> 		ls = new ArrayList<TomcatServlet>() ;
	final private List<TomcatConnectorBean> lc = new ArrayList<TomcatConnectorBean>() ;
	
	final private Map<String,String> mm = new TreeMap<String,String>() ;
	
	private String bdir ;
	private String hroo ;
	private String wlcm ;
	private String jvmr ;
	private String ldir ;
	private String llvl ;
	
	private Embedded embd ;
	private Engine   engn ;
	private Host     host ;
	private Context  rctx ;
	
	final public String getBaseDirectory()		{ return bdir ; }
	final public String getHttpRoot()			{ return hroo ; }
	final public String getWelcomeFile()		{ return wlcm ; }
	final public String getJVMRoute()			{ return jvmr ; }
	
	final public String getLoggingDirectory() 	{ return ldir ; }
	final public String getLoggingLevel()		{ return llvl ; }
	
	final public List<TomcatServlet> getServlets()			{ return ls ; }
	final public List<TomcatConnectorBean> getConnectors()	{ return lc ; }
	
	final public Map<String,String> getMimeMapping()	{ return mm ; }
	
	@Export
	@MethodInfo(desc="Adds the specified connector to the Tomcat engine")
	final public void addConnector(
		@ArgInfo(name="connector", desc="A instance of TomcatConnectorBean") TomcatConnectorBean tcb)
	{ lc.add(tcb); }
	
	@SuppressWarnings("unchecked")
	@MethodInfo(desc="Initializer contract implementation. Called at the framework initialization. Must not be called by others means")
	final public void postCreate(Element para)
	throws Exception 
	{
		// engine
		Element engnLmen = JDOMUtil.getElement(para, ConstTomcat.ENGINE) ;
		bdir  = engnLmen.getAttributeValue(ConstTomcat.ENGINE_BDIR, ConstTomcat.defaultBaseDir) ;
		hroo = engnLmen.getAttributeValue(ConstTomcat.ENGINE_HROO, ConstTomcat.defaultHttpRoot) ;
		wlcm = engnLmen.getAttributeValue(ConstTomcat.ENGINE_WELC, ConstTomcat.defaultWelcomeFile) ;
		jvmr = engnLmen.getAttributeValue(ConstTomcat.ENGINE_JVMR, ConstTomcat.defaultJvmRoute) ; 
		
		// logging
		Element loggLmen = JDOMUtil.getElement(engnLmen, ConstTomcat.LOGGING) ;
		ldir  = loggLmen.getAttributeValue(ConstTomcat.LOGGING_DIR, ConstTomcat.defaultLoggingDirectory) ;
		llvl = loggLmen.getAttributeValue(ConstTomcat.LOGGING_LEVL, ConstTomcat.defaultLoggingLevel) ;
		
		// servlet
		List<Element> svts = engnLmen.getChildren(ConstTomcat.SERVLET) ;
		for(Element svt : svts)
		{
			boolean enab = Boolean.valueOf(svt.getAttributeValue(ConstTomcat.SERVLET_ENAB, Boolean.TRUE.toString())).booleanValue() ;
			if(!enab) continue ;
			
			try
			{ ls.add(TomcatServlet.create(svt)) ; }
			catch(Exception ex)
			{ Log.warning("Unable to configure servlet ("+svt.getAttributeValue(ConstTomcat.SERVLET_NAME)+")") ; }
		}
		
		// mime types
		mm.putAll(PContext.getDefaultMimeTypes()) ;
		
		List<Element> mimes = engnLmen.getChildren(ConstTomcat.MIME) ;
		for(Element mime : mimes)
		{
			String exte = mime.getAttributeValue(ConstTomcat.MIME_EXTE);
			String type = mime.getAttributeValue(ConstTomcat.MIME_TYPE) ;
			
			Object oldv = mm.put(exte, type) ;
			
			if(oldv!=null) Log.warning("Override mime type ("+exte+") old("+oldv+") new("+type+")") ;
		}			
	}
	
	@MethodInfo(desc="The contract implementation of the application portal aspect (AppPortal) for the Console application")
	final public void console(HttpServlet srv, HttpServletRequest req, HttpServletResponse res)
	throws Exception
	{
		TomcatEngineConsole.service(this, srv, req, res) ; 
	}
	
	final public void startup()
	throws Exception
	{
	 try{
		Log.info("Tomcat starting ....") ;
		
    	// Paths
		File home = new File(bdir) ;
		if(!home.exists())
		{	Log.severe("Tomcat server home directory does not exists ("+home.getAbsolutePath()+")") ;
			throw new FileNotFoundException(home.getAbsolutePath()) ; }
		
		Log.info("Tomcat home directory ("+home.getCanonicalPath()+")") ;
		
    	File wdir = new File(hroo) ;
    	if(!wdir.isAbsolute()) wdir = new File(home, hroo) ;
    	if(!wdir.exists())
    	{	Log.severe("HTTP root directory does not exists ("+wdir.getAbsolutePath()+")") ;
    		throw new FileNotFoundException(wdir.getAbsolutePath()) ; }
    	
    	Log.info("Tomcat web directory ("+wdir.getCanonicalPath()+")") ;
    	
    	// Configure tomcat logging
    	File logs = new File(ldir) ;
    	if(!logs.isAbsolute()) logs = new File(home, ldir) ;
    	
    	PTomcatLog.configure(logs.getCanonicalPath(), llvl) ;
    	
    	// Service initialization
    	embd = new Embedded() ;
    	embd.setName("Azuki framework/"+Framework.getVersion()) ;
    	
    	embd.setCatalinaHome(bdir) ;
    	embd.setCatalinaBase(bdir) ;
    	
    	/*
    	  I prevented Tomcat from using creating its own JNDI Context 
    	  (and having new InitialContext() resolve to a preexisting Context),
    	  by setting the StandardContext.setUseNaming(false),
    	  which can be done in code or I guess through the context.xml file as well. */   	
    	/* Bob : je pense que si on met true alors on recupere le JNDI tomcat 
    	 * lors d'une requete HTTP */
    	embd.setUseNaming(false) ;	
    	
		// Requests engine
		engn = embd.createEngine() ;
		engn.setName("tomcat-embedded") ;
		engn.setDefaultHost(jvmr) ;
		engn.setJvmRoute(jvmr) ;	// AJP13 Load-Balancing	
	
		// Virtual host : "./web"
    	host = embd.createHost(jvmr, wdir.getCanonicalPath()) ;
    	host.setParentClassLoader(TomcatEngineBean.class.getClassLoader()) ; // otherwise some catalina classes not found !
    	
    	engn.addChild(host) ;
    	
    	// Root context initialisation
    	rctx = PContext.initRootContext(embd, host.getAppBase(), wlcm, ls, mm) ;
    	
    	host.addChild(rctx) ;
    	
		// Register Engine
		embd.addEngine(engn) ;
		
    	// Adding connectors
		PConnector.attach(lc, embd) ;
		
		try
     	{
     		embd.start() ;
     	}
        catch (LifecycleException e)
		{
        	Log.severe(Log.getStack(e)) ;
            if(e.getMessage().indexOf(BindException.class.getName())!=-1)
            	forceStart(embd.findConnectors(), e) ;
        }
	}
    catch(Exception ex)
    {
    	Log.severe(ex) ;
    	throw ex ;
    }
	}	
	
	// force the connectors following the first with errors to start
	final static private void forceStart(Connector[] cs, LifecycleException lce)
	{
		boolean first = true ;
		int fails = 0 ;
		
		for(int i=0 ; i<cs.length ; i++)
        {
        	Connector c = cs[i] ;
        	if(!c.isAvailable())
        	{
        		if(!first)
        		{
        			try
        			{
        				lce = null ;
        				c.start() ;
        			}
        			catch(LifecycleException e)
        			{ lce = e ; }
        		}
        		first = false ;
        		
        		if(lce!=null)
        		{
        			String err = c.getScheme()+"-"+c.getPort()+" is not available ! ("+lce.getLocalizedMessage()+")" ; 
        			Log.severe(err) ;
        			fails++ ;
        		}
        	}
        }
		
		// aucun connecteur de demarre !
		//if(fails==cs.length) System.exit(-1) ;
	}
}