/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.beandoc ;

import java.io.InputStream;
import java.lang.reflect.*;
import java.util.*;

import org.azuki.common.consts.*;
import org.azuki.common.framework.*;
import org.azuki.common.reflect.*;
import org.azuki.common.util.*;
import org.azuki.framework.aop.IContract;
import org.jdom.*;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

public class BeanDocumentationParser
{
	// XML namespaces
	final static private Namespace xsi = Namespace.getNamespace("xsi", "http://www.w3.org/2001/XMLSchema-instance") ;
	
	// XML elements & attributes
	final static private String BEANDOC = "beandoc" ;
	
	final static private String DEFN = "definition" ;
	final static private String DEFN_NAME = "name" ;
	final static private String DEFN_JARF = "jarfile" ;
	final static private String DEFN_CLAS = "class" ;
	final static private String DEFN_VERS = "version" ;
	final static private String DEFN_DESC = "description" ;
	final static private String DEFN_GRUP = "group" ;
	final static private String DEFN_ICON = "icon" ;
	
	final static private String ASPT = "aspect" ;
	final static private String ASPT_METH = "methodselection" ;
	final static private String ASPT_FILD = "fieldselection" ;
	
	final static private String EXPM = "exportmethod" ;
	final static private String METH = "method" ;
	final static private String METH_STUR = "signature" ;
	final static private String METH_DESC = "description" ;
	
	final static private String MPXY = "methodproxy" ;
	final static private String FILD = "field" ;
	final static private String FILD_STUR = "signature" ;
	final static private String FILD_DESC = "description" ;
	final static private String FILD_MDTY = "mandatory" ;
	
	final static private String PCUT = "defaultpointcut" ;
	final static private String BEAN = "bean" ;
	final static private String BEAN_NAME = "name" ;
	final static private String BEAN_DESC = "description" ;
	
	final static private String DOCU = "documentation" ;
	
	final static public Document parse(BeanDefinition bdef)
	throws Exception
	{
		Element defn = buildDefinition(bdef) ;
		
		if(bdef.isAspect()) defn.addContent(buildAspectDefinition(bdef)) ;
		
		defn.addContent(buildExportMethods(bdef)) ;
		defn.addContent(buildMethodProxies(bdef)) ;
		defn.addContent(buildDefaultPointcuts(bdef)) ;
		
		Element docu = buildDocumentation(bdef) ;
		
		Element root = new Element(BEANDOC) ;
		root.addNamespaceDeclaration(xsi) ;
		root.addContent(defn) ;
		root.addContent(docu) ;
		
		//System.out.println(JDOMUtil.xmlout.outputString(root)) ;
		
		return new Document(root) ;
	}
	
	final static private Element buildDefinition(BeanDefinition bdef)
	{
		Element lmen = new Element(DEFN) ;
		lmen.setAttribute(DEFN_NAME, bdef.getName()) ;
		lmen.setAttribute(DEFN_JARF, bdef.getJarFile()) ;
		lmen.setAttribute(DEFN_CLAS, bdef.getBeanClass().getName()) ;
		lmen.setAttribute(DEFN_VERS, bdef.getVersion()) ;
		
		String desc = bdef.getDescription() ;
		if(StringUtil.isEmpty(desc)) desc = "No description" ;
		lmen.setAttribute(DEFN_DESC, desc) ;
		
		String grup = bdef.getGroup() ;
		if(StringUtil.isEmpty(grup)) grup = "none" ;
		lmen.setAttribute(DEFN_GRUP, grup) ;
		
		String icon = bdef.getIcon() ;
		if(StringUtil.isEmpty(icon)) icon = "none" ;
		lmen.setAttribute(DEFN_ICON, icon) ;
		
		return lmen ;
	}
	
	final static private Element buildAspectDefinition(BeanDefinition bdef)
	{
		Element lmen = new Element(ASPT) ;
		
		AspectDefinition adef = bdef.getAspectDefinition() ;
		if(adef==null) return lmen ;
		
		IContract ctrt = adef.getContractObject() ;
		if(ctrt==null) return lmen ; //FIXME: definition is not valid...
		
		String msel = SelectionUtil.asString(ctrt.getMethodSelection()) ;
		String fsel = SelectionUtil.asString(ctrt.getFieldSelection()) ;
		
		lmen.setAttribute(ASPT_METH, msel) ;
		lmen.setAttribute(ASPT_FILD, fsel) ;
		
		return lmen ;
	}
	
	final static private Element buildExportMethods(BeanDefinition bdef)
	{
		Element lmen = new Element(EXPM) ;
		
		Collection<Method> ms = BeanReflect.getMethodsExported(bdef.getBeanClass()) ;
		
		for(Method m : ms)
		{
			Signature stur = new Signature(m) ;
			
			Element curr = new Element(METH) ;
			curr.setAttribute(METH_STUR, stur.toString()) ;
			curr.setAttribute(METH_DESC, Annotations.getMethodDescription(m)) ;
			
			lmen.addContent(curr) ;
		}
		
		return lmen ;
	}
	
	final static private Element buildMethodProxies(BeanDefinition bdef)
	{
		Element lmen = new Element(MPXY) ;
		
		Field[] fs = BeanReflect.getMethodProxyFields(bdef.getBeanClass()) ;
		
		for(Field f : fs)
		{
			String  stur = BeanReflect.getMethodProxySignature(f) ;
			String  desc = Annotations.getDependencyInjectionDescription(f) ;
			boolean mdty = Annotations.isDependencyInjectionMandatory(f) ;
			
			Element curr = new Element(FILD) ;
			curr.setAttribute(FILD_STUR, stur) ;
			curr.setAttribute(FILD_DESC, desc) ;
			curr.setAttribute(FILD_MDTY, String.valueOf(mdty)) ;
			
			lmen.addContent(curr) ;
		}
		
		return lmen ;
	}
	
	final static private Element buildDefaultPointcuts(BeanDefinition bdef)
	{
		Element lmen = new Element(PCUT) ;
		
		List<Element> es = JDOMUtil.listChildren(bdef.getElement(), Tag.POINTCUT) ;
		if(es.isEmpty()) return lmen ;
		
		for(Element e : es)
		{
			String name = e.getAttributeValue(Attr.NAME) ;
			String labl = e.getAttributeValue(Attr.LABL) ;
			
			BeanDefinition pcdf = bdef.getRepository().find(name) ;
			String desc = pcdf==null ? "Bean not found" : pcdf.getDescription() ;
			
			Element curr = new Element(BEAN) ;
			curr.setAttribute(BEAN_NAME, Util.oneString(name, labl)) ;
			curr.setAttribute(BEAN_DESC, desc) ;
			
			lmen.addContent(curr) ;
		}
		
		return lmen ;
	}
	
	final static private Element buildDocumentation(BeanDefinition bdef)
	{
		Element lmen = new Element(DOCU) ;
		
		String bdoc = bdef.getDocumentation() ;
		if(StringUtil.isEmpty(bdoc)) return lmen ;
		
		Document xdoc = null ;
		
		InputStream is = null ;
		try
		{
			is = bdef.getBeanClass().getClassLoader().getResourceAsStream(bdoc) ;
			
			xdoc = JDOMUtil.saxBuilder.build(is) ;
		}
		catch(Exception ex)
		{}
		finally
		{
			if(is!=null) try{ is.close() ; } catch(Exception ex){}
		}
		
		if(xdoc==null) return lmen ;
		
		Element root = xdoc.getRootElement() ;
		if(root!=null) lmen.addContent(root.cloneContent()) ;
		
		return lmen;
	}
}