/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.console ;

import java.io.*;
import java.util.*;

import org.azuki.common.util.StringUtil;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

final public class Template
{
	final protected LinkedList<Part> parts = new LinkedList<Part>() ;
	
	public Template(String text)
	throws IOException
	{
		this(new StringReader(text)) ;
	}
	
	public Template(InputStream is)
	throws IOException
	{
		this(new InputStreamReader(is)) ;
	}
	
	public Template(Reader reader)
	throws IOException
	{
		BufferedReader br = new BufferedReader(reader) ;
		
		StringBuilder bdr = new StringBuilder() ;
		
		while(true)
		{
			String line = br.readLine() ;
			if(line==null) break ;
			
			if(line.startsWith("@TEMPLATE:END-PART"))
			{
				String partName = partName(line) ;
				if(partName!=null)
				{	parts.add(new Part(partName, bdr.toString().trim())) ;
					bdr = new StringBuilder() ; }
					
				continue ;
			}
			
			bdr.append(line).append('\n') ;
		}
		
		parts.add(new Part(null, bdr.toString())) ;
		
		try { if(br!=null) br.close() ; }
		catch(Exception ex){}
	}
	
	final static private String partName(String s)
	{
		int idx1 = s.indexOf('(') ;
		if(idx1==-1) return null ;
		
		int idx2 = s.indexOf(')') ;
		if(idx1==-1) return null ;
		
		return s.substring(idx1+1, idx2) ;
	}
	
	final public Part getPart(String name)
	{
		if(name==null) return null; 
		
		for(Part part : parts)
			if(name.equals(part.getName())) return part ;
		
		return null ;
	}
	
	final public Part getFirst()
	{
		return parts.getFirst() ;
	}
	
	final public Part getLast()
	{
		return parts.getLast() ;
	}
	
	final public void replace(String oldStr, String newStr)
	{
		for(Part part : parts)
			part.replace(oldStr, newStr) ;
	}
	
	final public List<Part> getParts()
	{
		return Collections.unmodifiableList(parts) ;
	}
	
	/*
	 * Object
	 */
	@Override
	public String toString()
	{
		StringBuilder bdr = new StringBuilder() ;
		for(Part part : parts) bdr.append(part) ;
		return bdr.toString() ;
	}
	
	/**
	 * Template part
	 */
	final static public class Part
	{
		final private String name ;
		private String content ;
		
		Part(String name, String content)
		{
			this.name = name ;
			this.content = content ;
		}
		
		final public String getName()
		{
			return name ;
		}
		
		final public void replace(String oldStr, String newStr)
		{
			this.content = StringUtil.replace(content, oldStr, newStr) ;
		}
		
		@Override
		public String toString()
		{
			return content ;
		}
	}
}