/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.console.html ;

import java.util.*;

import org.azuki.common.console.html.renderer.TableCellRenderer;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

final public class Table extends Element
{	
	public Table()
	{
		super("table") ;
	}
	
	public Table(List<List> data)
	{
		this() ;		
		for(List list : data) addRow(list) ;
	}
	
	public Table(Object[][] data)
	{
		this() ;
		for(Object arr : data) addRow(arr) ;
	}
	
	final private Headers getHeadersImpl()
	{
		List<Headers> tmp = getContents(Headers.class) ;
		return tmp.isEmpty() ? null : tmp.get(0) ;
	}
	
	final public void setHeaders(List<Header> headers)
	{
		Headers headImpl = getHeadersImpl() ;
		if(headImpl==null)
		{	headImpl = new Headers() ;
			addContent(0, headImpl) ; }
		
		headImpl.setHeaders(headers) ;
	}
	
	final public void setHeaderNames(List list)
	{
		setHeaderNames(list.toArray()) ;
	}
	
	final public void setHeaderNames(Object... arr)
	{
		Headers headImpl = getHeadersImpl() ;
		if(headImpl==null)
		{	headImpl = new Headers() ;
			addContent(0, headImpl) ; }
		
		headImpl.clearContents() ;
		
		for(Object o : arr) headImpl.addHeader(o) ;
	}
	
	final public int headerCount()
	{
		return getHeaders().size() ;
	}
	
	final public Header getHeader(int index)
	{
		return getHeaders().get(index) ;
	}
	
	@SuppressWarnings("unchecked")
	final public List<Header> getHeaders()
	{
		Headers headImpl = getHeadersImpl() ;
		return headImpl==null ? Collections.EMPTY_LIST : headImpl.getHeaders() ;
	}
	
	
	final public void addRow(Row row)
	{
		addContent(row) ;
	}
	
	final public void addRow(List list)
	{
		Row row = new Row() ;
		for(Object o : list) row.addCell(o) ;
		addRow(row) ;
	}
	
	final public void addRow(Object... arr)
	{
		Row row = new Row() ;
		for(Object o : arr) row.addCell(o) ;
		addRow(row) ;
	}
	
	final public int rowCount()
	{
		return getRows().size() ;
	}
	
	final public Row getRow(int index)
	{
		return getRows().get(index) ;
	}
	
	final public List<Row> getRows()
	{
		return getContents(Row.class) ;
	}
	
	final public void setColumnRenderer(int index, TableCellRenderer cellRenderer)
	{
		for(Row row : getRows())
		{
			if(index<row.size())
				row.getCell(index).setCellRenderer(cellRenderer) ;
		}
	}
	
	final static public class Row extends Element
	{
		public Row()
		{
			super("tr") ;
		}
		
		public Row(List<Cell> c)
		{
			this() ;
			setContents(c) ;
		}
		
		final public void addCell(Object obj)
		{
			addCell(new Cell(obj)) ;
		}
		
		final public void addCell(Cell cell)
		{
			addContent(cell) ;
		}
		
		final public int size()
		{
			return getCells().size() ;
		}
		
		final public Cell getCell(int index)
		{
			return getCells().get(index) ;
		}
		
		final public List<Cell> getCells()
		{
			return getContents(Cell.class) ;
		}
	}
	
	final static private class Headers extends Element
	{
		public Headers()
		{
			super("tr") ;
		}
		
		public Headers(List<Header> headers)
		{
			this() ;
			setContents(headers) ;
		}
		
		final public void addHeader(Object obj)
		{
			addHeader(new Header(obj)) ;
		}
		
		final public void addHeader(Header header)
		{
			addContent(header) ;
		}
		
		final public int size()
		{
			return getHeaders().size() ;
		}
		
		final public Header getHeader(int index)
		{
			return getHeaders().get(index) ;
		}
		
		final public List<Header> getHeaders()
		{
			return getContents(Header.class) ;
		}
		
		final public void setHeaders(List<Header> headers)
		{
			setContents(headers) ;
		}
	}
	
	final static public class Header extends Element
	{
		public Header()
		{
			super("th") ;
		}
		
		public Header(Object obj)
		{
			this() ;
			setContent(new Text(obj)) ;
		}
	}
	
	final static public class Cell extends Element
	{
		private TableCellRenderer cellRenderer ;
		
		Cell(String tag)
		{
			super(tag) ;
		}
		
		public Cell()
		{
			this("td") ;
		}
		
		public Cell(Object obj)
		{
			this() ;
			setContent(new Text(obj)) ;
		}
		
		final public void setCellRenderer(TableCellRenderer cellRenderer)
		{
			this.cellRenderer = cellRenderer ;
		}
		
		@SuppressWarnings("unchecked")
		@Override
		final public String toString()
		{
			if(cellRenderer==null) return super.toString() ;
			
			StringBuilder bdr = new StringBuilder() ;
			
			bdr.append('<').append(name).append(' ') ;
			for(Attribute attr : getAttributes())
				bdr.append(attr).append(' ') ;
			
			if(getContents().isEmpty())
			{	bdr.append("/>").append('\n') ;
				return bdr.toString() ; }
			
			bdr.append('>').append('\n') ;
			
			for(Content c : getContents())
			{
				if(c instanceof Text) bdr.append(cellRenderer.render(((Text)c).getValue())) ;
				else 				  bdr.append(c) ;
			}
				
			
			bdr.append("</").append(name).append('>').append('\n') ;
			
			return bdr.toString() ;
		}
	}
}