/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.console.http ;

import java.io.*;
import java.net.URLDecoder;
import java.util.*;

import javax.servlet.http.*;

import org.apache.commons.codec.binary.Base64;
import org.azuki.common.console.http.HttpHeader;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

final public class HttpRequest
{
	final static public String HEADER_USER_AGENT = "User-Agent" ;
	
	final static public boolean isIE(HttpServletRequest req)
	{
		String ua = req.getHeader(HEADER_USER_AGENT) ;
		return ua==null ? false : ua.indexOf("MSIE")!=-1 ;
	}
	
	// Return parameter value
	final static public String get(HttpServletRequest requ, String para)
	{
		return requ.getParameter(para) ; 
	}
	
	// Return parameter value
	// Return default value if not exists
	final static public String get(HttpServletRequest requ, String para, String def)
	{
		String val = requ.getParameter(para) ;
		return val==null ? def : val ;
	}
	
	// Return parameter value
	// Return default value if not exists or empty
	final static public String get(HttpServletRequest requ, String para, String def, boolean emptyIsNull)
	{
		String val = get(requ, para, def) ;
		if(!emptyIsNull) return val ; 
		
		return val==null||val.trim().isEmpty() ? def : val ;
	}

	
	// Return the cookie value
	final static public String getCookie(HttpServletRequest req, String cookieName)
	{
		Cookie[] cookies = req.getCookies() ;
		if(cookies==null) return null ;
		
		for(Cookie cookie : cookies)
		{	if(cookie.getName().equals(cookieName))
				return cookie.getValue() ; }
		
		return null ;
	}
	
	// Return the cookie values
	final static public Collection<String> getCookies(HttpServletRequest req, String cookieName)
	{
		Cookie[] cookies = req.getCookies() ;
		if(cookies==null) return null ;
		
		ArrayList<String> arr = new ArrayList<String>() ;
		
		for(Cookie cookie : cookies)
		{	if(cookie.getName().equals(cookieName))
				 arr.add(cookie.getValue()) ; }
		
		return arr ;
	}
	
	// Return the "well-decoded" request URI
	final static public String decodeURI(HttpServletRequest req)
	{
		String enc = req.getCharacterEncoding() ;
		if(enc==null) enc = System.getProperty("file.encoding") ;
		
		String ruri = req.getRequestURI() ;
		try
		{ ruri = URLDecoder.decode(ruri, enc) ; }
		catch(UnsupportedEncodingException uee)
		{ uee.printStackTrace() ; }
		
		return ruri ;
	}
	
	// Make URL for current HttpServletRequest with specified parameters
	final static public String mkURL(HttpServletRequest req, Map<String,String> params)
	{
		return mkURL(req, params, false) ;
	}
	
	final static public String mkURL(HttpServletRequest req, Map<String,String> params, boolean esc)
	{
		StringBuilder sb = new StringBuilder(req.getRequestURI()) ;
		if(params==null) return sb.toString() ;
		
		sb.append('?') ;
		
		boolean first = true ;
		
		for(Map.Entry<String,String> ety : params.entrySet())
		{
			String val = ety.getValue() ;
			if(val==null) continue ;
			
			if(!first) 	sb.append(esc ? "&amp;" : "&") ;
			else		first = false ;
			
			sb.append(ety.getKey()).append('=').append(val) ;
		}
		
		return sb.toString() ;
	}
	
	
	// Return request parameters map
	@SuppressWarnings("unchecked")
    final static public Map<String,String> getParameterMap(HttpServletRequest req)
	{
		Map<String,String> map = new HashMap<String,String>() ;
		
		Enumeration<String> enu = req.getParameterNames() ;
		while(enu.hasMoreElements())
		{
			String key = enu.nextElement() ;
			map.put(key, req.getParameter(key)) ;
		}
		
		return map ;
	}
	
	
	// Return request codebase (http://host[:port]/path)
	final static public String codebase(HttpServletRequest req)
	{
		StringBuffer url = req.getRequestURL() ;
		
		int idx = url.indexOf("?") ;
		if(idx!=-1) url.delete(idx, url.length()) ;
		
		idx = url.lastIndexOf("/") ;
		if(idx!=-1) url = url.delete(idx, url.length()) ;
		
		return url.toString() ;
	}
	
	
	// Return server base (http://host[:port])
	final static public String serverBase(HttpServletRequest req)
	{
		String uri = req.getRequestURI() ;
		
		StringBuffer url = req.getRequestURL() ;
		url.delete(url.indexOf(uri), url.length()) ;
		
		return url.toString() ;
	}
	
	// Return BASIC authenfification information 
	final static public HttpAuthInfo<String,String> getBasicAuth(HttpServletRequest req)
	{
		String auth = req.getHeader(HttpHeader.AUTHORIZATION) ;
		if(auth==null) return null ;
		
		int idx = auth.indexOf(' ') ;
		String typ = auth.substring(0, idx).trim() ;
		if(!typ.equalsIgnoreCase(HttpServletRequest.BASIC_AUTH))
			return null ;
		
		byte[] byts = auth.substring(idx+1).trim().getBytes() ;
		if(!Base64.isArrayByteBase64(byts))
			return null ;
		
		String dec = new String(Base64.decodeBase64(byts)) ;
		
		idx = dec.indexOf(':') ;
		if(idx==-1) return null ;
		
		return new HttpAuthInfo<String,String>(dec.substring(0, idx), dec.substring(idx+1)) ;
	}
	
	// Return request body content
	final static public byte[] getContent(HttpServletRequest req)
	{
		ByteArrayOutputStream baos = new ByteArrayOutputStream() ;
		
		InputStream is = null ;
		
		try
		{
			is = req.getInputStream() ;
			
			byte[] buf = new byte[4096] ;
			while(true)
			{
				int len = is.read(buf) ;
				if(len==-1) break ;
				baos.write(buf, 0, len) ;
			}
			
			baos.flush() ;
		}
		catch(Exception ex)
		{
			return null ;
		}
		finally
		{
			try
			{ if(is!=null) is.close() ; baos.close() ; }
			catch(Exception e){}
		}
		
		return baos.toByteArray() ;
	}
	
	// affiche le resultat d'appel de methode de l'API Servlet
	final static public void dumpAPI(HttpServletRequest req, PrintStream ps)
	{
		String dmp = dumpAPI(req) ;
		ps.println(dmp) ;
		ps.flush() ;
	}
	
	final static public String dumpAPI(HttpServletRequest req)
	{
		StringBuilder sb = new StringBuilder()
		.append("Scheme         : ").append(req.getScheme()).append('\n')
		.append("Path info      : ").append(req.getPathInfo()).append('\n')
		.append("Path translated: ").append(req.getPathTranslated()).append('\n')
		.append("Context path   : ").append(req.getContextPath()).append('\n')
		.append("Servlet path   : ").append(req.getServletPath()).append('\n')
		.append("URI            : ").append(req.getRequestURI()).append('\n')
		.append("URL            : ").append(req.getRequestURL()).append('\n')
		.append("Query string   : ").append(req.getQueryString());
		return sb.toString() ;
	}
	
	// affiche le contenu de la requete
	final static public void dump(HttpServletRequest req, PrintStream ps, boolean content)
	{
		String dmp = dump(req, content) ;
		ps.println(dmp) ;
		ps.flush() ;
	}
	final static public String dump(HttpServletRequest req, boolean content)
	{
		String meth = req.getMethod() ;
		String uri = req.getRequestURI() ;
		String prot = req.getProtocol() ;
		
		String quer = req.getQueryString() ;
		if(quer!=null && !quer.isEmpty())
			uri += "?"+quer ;
		
		StringBuffer sb = new StringBuffer()
		.append(meth).append(' ').append(uri).append(' ').append(prot) ;
		
		Enumeration hdrs = req.getHeaderNames() ;
		while(hdrs.hasMoreElements())
		{
			String head = (String)hdrs.nextElement() ;
			sb.append('\n').append(head).append(": ") ;
			
			Enumeration vals = req.getHeaders(head) ;
			while(vals.hasMoreElements())
			{
				sb.append(vals.nextElement()) ;
				if(vals.hasMoreElements()) sb.append("; ") ;
			}
		}
		
		if(content)
		{
			sb.append('\n')
			.append(new String(getContent(req))) ;
		}
		
		return sb.toString() ;
	}
}