/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.dag;

import java.util.*;

/**
 * 
 *  DAG : Directed Acyclic Graph
 * 
 * @author Robert Bakic (robert.bakic@gmail.com)
 */
public class DAG<T> 
{
	// Test
	final static public void main(String[] args)
	{
		DAG<Object> tp = new DAG<Object>() ;

		try
		{
			tp.addEdge("B", null, "A", null) ;
			
			tp.addEdge("C", null, "A", null) ;

			tp.addEdge("C", null, "B", null) ;
			tp.addEdge("D", null, "C", null) ;
			tp.addEdge("D", null, "B", null) ;
			tp.addEdge("F", null, "B", null) ;

//			tp.addEdge("A", null,"F", null) ; // add this line,  for cycling exception
			
			if(tp.hasCycle())
			{
				List<Node<Object>> ll = tp.getCycle() ; 
				tp.printNodes(ll, true) ;
			}
			else
			{
				List<Node<Object>> sorted = tp.sort() ;
				tp.printNodes(sorted, false) ;
			}
		}
		catch(Exception e)
		{
			e.printStackTrace() ;
		}
	}
	
	final private ArrayList<Link<T>> allinks = new ArrayList<Link<T>>() ;
	
	// parent=aspect, child=target
	final public void addEdge(String pkey, T parent, String ckey, T child)
	{
		allinks.add(new Link<T>(new Node<T>(pkey, parent), new Node<T>(ckey, child))) ;
	}

	// parent=aspect, child=target
	// return true if the link has been added, false otherwise
	final public boolean addEdgeIfNoCycle(String pkey, T parent, String ckey, T child)
	{
		Link<T> link = new Link<T>(new Node<T>(pkey, parent), new Node<T>(ckey, child)) ;
		allinks.add(link) ;
		if(hasCycle())
		{
			allinks.remove(link) ;
			return false ;
		}
		return true ;
	}

	@SuppressWarnings("unchecked")
	final public List<Node<T>> sort()
	throws Exception
	{
		// ll est la liste des liens, ne doit pas etre modifiée
		ArrayList<Link<T>> cpl = (ArrayList<Link<T>>)allinks.clone() ;		
	
		List<Node<T>> orphan = orphanNodes(cpl) ;
		List<Node<T>> sorted = new ArrayList<Node<T>>() ;
		
		while(!orphan.isEmpty())
		{
			Node<T> n = orphan.remove(0) ;
			sorted.add(n) ;
			
			List<Link> childs = getChildrenLink(n, cpl) ;
			
			for(Link e : childs)
			{
				cpl.remove(e) ;
				Node<T> m = e.child ;
				if(!hasParent(m, cpl)/* && !orphan.contains(m)*/)
					orphan.add(m) ;
			}
		}
	
		if(!cpl.isEmpty())
			throw new Exception("Cycling exception") ;
		
		return sorted ;
	}

	final public boolean hasCycle()
	{
		return !getCycle().isEmpty() ;
	}

	// false : {A, B}
	// true  : A  -> B 
	final private void printNodes(List<Node<T>> nodes, boolean format)
	{
		StringBuilder sb = new StringBuilder() ;
		
		if(!format) sb.append("{") ;
				
		boolean one = false ;
		for(Node node : nodes)
		{
			if(!one) one = true ;
			else
			{
				if(format) sb.append( " --> ") ;
				else	   sb.append(", ") ;
			}

			sb.append(node.getKey()) ;
		}
		
		if(!format) sb.append("}") ;
		
		System.out.println(sb.toString()) ;		
	}

	/*
	final private void printLinks(List<Link<T>> list)
	{
		for(Link l : list)
			System.out.println("("+l+") : "+l.child + " < " + l.parent) ;				
	}
	*/
	
	final public List<Node<T>> getCycle()
	{
		ArrayList<Node<T>> nodes = allParentNodes(allinks) ;
		
		// Destine a contenir la reponse
		ArrayList<Node<T>> cycle = new ArrayList<Node<T>>() ;

		for(Node<T> node : nodes) // Pour chaque nodes
		{
			if(chain(node, new ArrayList<Node<T>>(), cycle)) return cycle ;	
		}
		
		return  cycle;
	}	
	
	@SuppressWarnings("unchecked")
	final public boolean chain(Node<T> node, ArrayList<Node<T>> path, ArrayList<Node<T>> cycle)
	{
		if(path.contains(node))
		{
			cycle.addAll(path) ;
			cycle.add(node) ; // deux fois element
			// printNodes(cycle, true) ;
			return true ; // cycle trouve
		}
		
		path.add(node) ;
		
		List<Node<T>> childs =  getChildrenNodes(node) ;
		for(Node<T> child : childs)
		{
			// clone du path courant qui est bon jusque là
			ArrayList<Node<T>> pcopy = (ArrayList<Node<T>>)path.clone() ;
		
			boolean cy = chain(child, pcopy, cycle) ;
			
			if(cy) return true ; // remonter cycling error
		}
		
		return false ;
	}
	
	// Verify qu'une intance a un parent dans la liste
	final private boolean hasParent(Node<T> node, List<Link<T>> list)
	{
		for(Link l : list)
			if(node.equals(l.child)) return true ;
		return false ;
	}
	
	final private List<Link> getChildrenLink(Node<T> parent, List<Link<T>> list)
	{
		List<Link> cl = new ArrayList<Link>() ;
		
		for(Link l : list)
			if(parent.equals(l.parent)) cl.add(l) ;
		
		return cl ;
	}

	@SuppressWarnings("unchecked")

	final private List<Node<T>> getChildrenNodes(Node<T> parent)
	{
		List<Node<T>> childs = new ArrayList<Node<T>>() ;
		
		for(Link l : allinks)
			if(parent.equals(l.parent)) childs.add(l.child) ;
		
		return childs;
	}

	
	// List des instance sans parents (sans apsect)
	final private ArrayList<Node<T>> orphanNodes(List<Link<T>> list)
	{
		ArrayList<Node<T>> orphan = new ArrayList<Node<T>>() ;
		
		for(Link<T> l : list)
		{
			Node<T> p = l.parent ;
			boolean hp = hasParent(p, list) ;
			if(!hp && !orphan.contains(p)) 
				orphan.add(p) ;
		}
		
		return orphan ;
	}

	final private ArrayList<Node<T>> allParentNodes(List<Link<T>> list)
	{
		ArrayList<Node<T>> all = new ArrayList<Node<T>>() ;
		
		for(Link<T> l : list)
		{
			Node<T> p = l.parent ;
			if(!all.contains(p)) all.add(p) ;
		}
		
		return all ;
	}
}
