/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.editor.jdom ;

import java.util.*;

import javax.swing.event.*;
import javax.swing.table.TableModel;

import org.jdom.Element;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

public class JDOMElementTableModel implements TableModel
{
	final protected EventListenerList listenerList = new EventListenerList() ;
	
	final protected Element lmen ;
	final protected String childName ;
	final protected List<String> childAttrs ;
	final protected Map<String,String> defValus ;
	
	protected JDOMElementTableModel()
	{
		this(new Element("root"), null, new ArrayList<String>()) ;
	}
	
	public JDOMElementTableModel(Element lmen, String childName, String... childAttrs)
	{
		this(lmen, childName, Arrays.asList(childAttrs), null) ;
	}
	
	public JDOMElementTableModel(Element lmen, String childName, List<String> childAttrs)
	{
		this(lmen, childName, childAttrs, null) ;	
	}
	
	public JDOMElementTableModel(Element lmen, String childName, String[] childAttrs, Map<String,String> defValus)
	{
		this(lmen, childName, Arrays.asList(childAttrs), defValus) ;
	}
	
	public JDOMElementTableModel(Element lmen, String childName, List<String> childAttrs, Map<String,String> defValus)
	{
		this.lmen = lmen ;
		this.childName = childName ;
		this.childAttrs = childAttrs ;
		this.defValus = defValus ;
	}
	
	final public Element getElement()
	{
		return lmen ;
	}
	
	final public String getChildName()
	{
		return childName ;
	}
	
	final public List<String> getChildAttributeNames()
	{
		return Collections.unmodifiableList(childAttrs) ;
	}
	
	@SuppressWarnings("unchecked")
	final public List<Element> getElements()
	{
		return lmen.getChildren(childName) ;
	}
	
	final public Element getElementAt(int rowIndex)
	{
		return getElements().get(rowIndex) ;
	}
	
	final public void newElement()
	{
		int index = getElements().size() ;
		newElement(index) ;
	}
	
	final public void newElement(int index)
	{
		Element lmen = new Element(childName) ;
		for(String attr : childAttrs)
		{
			String valu = defaultValueFor(attr, "") ;
			lmen.setAttribute(attr, valu) ;
		}
		
		addElement(index, lmen) ;
	}
	
	final public void addElement(int index, Element lmen)
	{
		getElements().add(index, lmen) ;
		
		fireTableRowsInserted(index, index) ;
	}
	
	final public void addElement(Element lmen)
	{
		getElements().add(lmen) ;
		
		int index = getElements().size() ;
		fireTableRowsInserted(index, index) ;
	}
	
	final public void removeElement(int rowIndex)
	{
		getElements().remove(rowIndex) ;
		fireTableRowsDeleted(rowIndex, rowIndex) ;
	}
	
	final protected String defaultValueFor(String attr, String nulv)
	{
		if(defValus==null) return nulv ;
		
		String dval = defValus.get(attr) ;
		return dval==null ? nulv : dval ;
	}
	
	/*
	 * Events
	 */
	public void fireTableDataChanged()
	{
		fireTableChanged(new TableModelEvent(this)) ;
	}

	public void fireTableStructureChanged()
	{
		fireTableChanged(new TableModelEvent(this, TableModelEvent.HEADER_ROW)) ;
	}

	public void fireTableRowsInserted(int firstRow, int lastRow)
	{
		fireTableChanged(new TableModelEvent(this, firstRow, lastRow, TableModelEvent.ALL_COLUMNS, TableModelEvent.INSERT)) ;
	}

	public void fireTableRowsUpdated(int firstRow, int lastRow)
	{
		fireTableChanged(new TableModelEvent(this, firstRow, lastRow, TableModelEvent.ALL_COLUMNS, TableModelEvent.UPDATE)) ;
	}

	public void fireTableRowsDeleted(int firstRow, int lastRow)
	{
		fireTableChanged(new TableModelEvent(this, firstRow, lastRow, TableModelEvent.ALL_COLUMNS, TableModelEvent.DELETE)) ;
	}

	public void fireTableCellUpdated(int row, int column)
	{
		fireTableChanged(new TableModelEvent(this, row, row, column)) ;
	}

	public void fireTableChanged(TableModelEvent e)
	{
		TableModelListener[] arr = listenerList.getListeners(TableModelListener.class) ;
		for(TableModelListener tml : arr)
			tml.tableChanged(e) ;
	}
	
	/*
	 * TableModel
	 */
	public int getRowCount()
	{
		return getElements().size() ;
	}
	
	public int getColumnCount()
	{
		return childAttrs.size() ;
	}

	public String getColumnName(int columnIndex)
	{
		return childAttrs.get(columnIndex) ;
	}
	
	public Class<?> getColumnClass(int columnIndex)
	{
		return String.class ;	//FIXME ??
	}

	public Object getValueAt(int rowIndex, int columnIndex)
	{
		String attrName = childAttrs.get(columnIndex) ;
		
		String attrValu = getElementAt(rowIndex).getAttributeValue(attrName) ;
		if(attrValu==null)  attrValu = defaultValueFor(attrName, null) ;
		
		return attrValu ;
	}

	public boolean isCellEditable(int rowIndex, int columnIndex)
	{
		return true ;
	}

	public void setValueAt(Object aValue, int rowIndex, int columnIndex)
	{
		getElementAt(rowIndex).setAttribute(childAttrs.get(columnIndex), String.valueOf(aValue)) ;
		fireTableCellUpdated(rowIndex, columnIndex) ;
	}

	public void addTableModelListener(TableModelListener l)
	{
		listenerList.add(TableModelListener.class, l) ;
	}
	
	public void removeTableModelListener(TableModelListener l)
	{
		listenerList.remove(TableModelListener.class, l) ;
	}
}