/*
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.framework ;

import java.io.InputStream;
import java.net.URL;
import java.util.*;
import java.util.logging.Logger;

import org.azuki.common.consts.*;
import org.azuki.common.util.StringUtil;
import org.azuki.framework.aop.IAspect;
import org.jdom.*;
import org.jdom.input.SAXBuilder;

/**
 * @author Robert Bakic (robert.bakic@gmail.com)
 */

final public class BeanRepository
{
	static private BeanRepository inst = null ;
	final static public BeanRepository instance() { return inst; }
	
	final static public void init(ClassLoader cl, Logger lg)
	{
		if(inst!=null) return ;
		
		long time = System.currentTimeMillis() ;
		
		inst = new BeanRepository(cl, lg) ;
		inst.restore() ;
		
		System.out.println("Beans definition loaded in "+(System.currentTimeMillis()-time)+"ms") ;
	}
	
	final private Map<String,BeanDefinition> bds = new HashMap<String,BeanDefinition>() ;
	
	final private ClassLoader cl ;
	final private Logger lg ;
	
	public BeanRepository(ClassLoader cl, Logger lg)
	{
		this.cl = cl ;
		this.lg = lg ;
	}
	
	final public ClassLoader getClassLoader()
	{
		return cl ;
	}
	
	final public BeanDefinition find(String name)
	{
		return bds.get(name) ;
	}
	
	final public Collection<BeanDefinition> all()
	{
		return bds.values() ;
	}
	
	@SuppressWarnings("unchecked")
	final public void restore()
	{
		lg.info("<<<<<<<<<<< READING STATIC BEANS DEFINITION (beans.xml) >>>>>>>>>>>>>>>\n") ;

		try
		{
			SAXBuilder saxb = new SAXBuilder() ;
			saxb.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false) ;
			
			Enumeration<URL> enu = cl.getResources(Const.BEANS_FILE) ;
			while(enu.hasMoreElements())
			{
			    URL url = enu.nextElement() ;
			    lg.info("Parsing: "+url) ;
			    
			    String jarf = getJarFile(url) ;
			    
			    try
		    	{
			    	InputStream is = url.openStream() ; // lecture
			    	
			    	Document doc = saxb.build(is) ;		
		    		is.close() ;
		    		
			    	List<Element> ll = doc.getRootElement().getChildren(Tag.BEAN) ;
	
			    	for(Element lm : ll)
			    	{
			    		// control integrity
			    		try
			    		{ check(lm) ; }
			    		catch(Exception ex)
			    		{ lg.severe("ERROR: "+ex.getMessage()) ; continue ; }
			    		
			    		// adding jar file information
			    		//lm = (Element)lm.clone() ;
			    		
			    		// BOB : null value not permitted in JDOM
			    		if(jarf==null) jarf="" ;
			    		lm.setAttribute(Attr.JARF, jarf) ;
			    		
			    		// register bean's definition
			    		defineBean(lm) ;
			    	}
		    	}
		    	catch(Exception ex)
		    	{
				    lg.severe("ERROR : "+ex.getMessage()) ;
		    		ex.printStackTrace() ;
		    	}
		    	
		    	System.out.println() ;
			}
		}
		catch(Exception e)
		{
			e.printStackTrace() ;
		}
	}
	
	final private void defineBean(Element lmen)
	{
		String name = null ;
		
		try
		{
			name = lmen.getAttributeValue(Attr.NAME) ;	
			
			String clas = lmen.getAttributeValue(Attr.CLAS) ;
			String jarf = lmen.getAttributeValue(Attr.JARF) ;
			String vers = lmen.getAttributeValue(Attr.VERS) ;
			
			Class c = cl.loadClass(clas) ;
			
			BeanDefinition bd = new BeanDefinition(this, name, c, jarf, vers, lmen) ;
			
			if(IAspect.class.isAssignableFrom(c))
			{
				Element aspt = lmen.getChild(Tag.ASPECT) ;
				
				String ctrt = null ;
				String edit = null ;
				
				if(aspt!=null)
				{
					ctrt = aspt.getAttributeValue(Attr.CTRT) ;
					edit = aspt.getAttributeValue(Attr.EDIT) ;
				}
				
				AspectDefinition ad = new AspectDefinition(bd, ctrt, edit) ;
				
				bd.setAspectDefinition(ad) ;
			}
			
			bds.put(name, bd) ;
			
			lg.info("Definition OK ["+name+"]") ;
		}
		catch(Exception e)
		{
			lg.info("Definition ["+name+"]   ERROR: "+e.getMessage()) ;
		}
	}
	
	// Put the controls here
    final private void check(Element lmen)
    throws Exception
    {
		// checking bean's name
		String name = lmen.getAttributeValue(Attr.NAME) ;
    	if(StringUtil.isEmpty(name))
    		throw new Exception("Invalid bean name ("+name+")") ;
    	
    	BeanDefinition bd = bds.get(name) ;
    	if(bd!=null)
    		throw new Exception("Bean ("+name+") already defined in jar file ("+bd.getJarFile()+")") ;
		
    	// checking bean's class
    	String clas = lmen.getAttributeValue(Attr.CLAS) ;
    	if(StringUtil.isEmpty(clas))
    		throw new Exception("Bean ("+name+") undefined '"+Attr.CLAS+"' attribute") ;
    	
    	try
    	{ cl.loadClass(clas) ; }
    	catch(Exception ex)
    	{ throw new Exception("Bean ("+name+") class ("+clas+") not found") ; }
    	
    	// checking bean's version
    	String vers = lmen.getAttributeValue(Attr.VERS) ;
    	if(StringUtil.isEmpty(vers))
    		throw new Exception("Bean ("+name+") undefined '"+Attr.VERS+"' attribute") ;
    	
    	try
    	{ Float.parseFloat(vers) ; }
    	catch(Exception ex)
    	{ throw new Exception("Bean ("+name+") version ("+vers+") is not valid") ; }
    }
	
	final static private String getJarFile(URL url)
    {
    	if(url==null) return null ;
    	
    	String surl = url.getFile() ;
    	int idx = surl.indexOf('!') ;
    	return idx==-1 ? null : surl.substring(0, idx) ;
    }
}
