/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.reflect ;

import java.lang.reflect.*;
import java.util.*;

/**
 * @author Robert Bakic (robert.bakic@gmail.com)
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

final public class ReflectUtil
{
	final static private Class[] NOARGS = new Class[0] ; // Just for easier comparison
	final static private Class[] BOXED_CLASSES = { Boolean.class, Character.class, Byte.class, Short.class, Integer.class, Long.class, Float.class, Double.class, Void.class } ;
	
	final static public boolean isFieldExists(Class<?> clas, String fnam)
	{
		try{ clas.getField(fnam) ; }
		catch(Exception ex){ return false ; }
		return true ;
	}
	
	final static public Field[] getFields(Class clas, Class<?> fieldType)
	{
		List<Field> list = new ArrayList<Field>() ;
		
		try
		{
			for(Field field : clas.getFields())
				if(fieldType.isAssignableFrom(field.getType())) list.add(field) ;
		}
		catch(Throwable t)
		{
			t.printStackTrace() ;
		}
		
		return list.toArray(new Field[list.size()]) ;
	}
	
	final static public boolean isMethodExists(Class<?> clas, String mnam)
	{
		for(Method meth : clas.getMethods())
			if(meth.getName().equals(mnam)) return true ;
		return false ;
	}
	
	final static public Method getMethod(Class<?> clas, String name)
	throws NoSuchMethodException
	{
		for(Method meth : clas.getMethods())
			if(meth.getName().equals(name)) return meth ;
		
		throw new NoSuchMethodException("Method ("+clas.getName()+"."+name+")") ;
	}
		
	// retourne la signature de la methode
	final static public String getMethodSignature(Method method)
	{
		try
		{
			return new Signature(method).getSignature() ;
		}
		catch(Exception ex)
		{
			System.err.println("Cannot build method signature ("+method+")") ;
			ex.printStackTrace() ;
		}
		
		return null ;
	}
	
	final static public Collection<Signature> getSignatures(Method[] methods)
	{
		if(methods==null || methods.length==0) return Collections.emptyList() ;
		
		Set<Signature> sturs = new TreeSet<Signature>() ;
		for(Method method : methods)
		{
			try	
			{ sturs.add(new Signature(method)) ; }
			catch(Exception ex){}
		}
		
		return sturs ;
	}
	
	final static public Collection<Signature> getSignatures(Collection<Method> methods)
	{
		if(methods==null || methods.isEmpty()) return Collections.emptyList() ;
		
		List<Signature> sturList = new ArrayList<Signature>() ;
		for(Method method : methods)
		{
			try	
			{ sturList.add(new Signature(method)) ; }
			catch(Exception ex){}
		}
		
		return sturList ;
	}
	
	final static private String shortString(Type type)
	{
		String tstr = type.toString() ;
		return tstr.substring(tstr.lastIndexOf('.')+1) ;
	}
	
	final static public String toString(Type type)
	{
		// recupere/modifie de ParameterizedTypeImpl.toString()  
		if(!(type instanceof ParameterizedType))
		{
			if(type instanceof Class)
				return type==void.class ? "void" : ((Class)type).getSimpleName() ;	
			return shortString(type) ;		
		}
			
		ParameterizedType pType = (ParameterizedType)type ;
		Type[] pTypeArgs = pType.getActualTypeArguments() ;
		
		Type defType = pType.getOwnerType() ;
		if(defType==null) defType = pType.getRawType() ;
	
		StringBuilder sb = new StringBuilder() ;
		sb.append(defType instanceof Class ? ((Class)defType).getSimpleName() : shortString(defType)) ;
        
		if(pTypeArgs!=null && pTypeArgs.length>0)
		{
			sb.append("<");
			
			boolean first = true ;
			for(Type t : pTypeArgs)
			{
				if(first) 	first = false ;
				else		sb.append(",") ;
				
				sb.append(toString(t)) ;
			}
			sb.append(">");
		}

		return sb.toString();
	}
	
	final static public String toString(Class[] classes)
	{
		StringBuilder sb = new StringBuilder() ;
		
		boolean first = true ; 
		for(Class clas : classes)
		{
			if(first) 	first = false ;
			else		sb.append(",") ;
			
			sb.append(clas.getSimpleName()) ;
		}
		
		return sb.toString() ;
	}
	
	final static public String toString(Type[] types)
	{
		StringBuilder sb = new StringBuilder() ;
		
		boolean first = true ; 
		for(Type type : types)
		{
			if(first) 	first = false ;
			else		sb.append(",") ;
			
			sb.append(toString(type)) ;
		}
		
		return sb.toString() ;
	}
	
	// retourne la liste des noms de champ/methode
	final static public <T extends Member> List<String> stringList(T[] ts)
	{
		if(ts==null) return Collections.emptyList() ;
		
		List<String> list = new ArrayList<String>() ;
		for(T t : ts) list.add(t.getName()) ;
		
		return list ;
	}
	
	final static public String getFieldDocumentation(Class<?> clas, String fieldName)
	{
		Field field = null ;
		try
		{ field = clas.getField(fieldName) ; }
		catch(Exception ex)
		{ /*ex.printStackTrace() ;*/ }
		
		StringBuilder bdr = new StringBuilder() ;
		bdr.append("<html>") ;
		bdr.append("<b>").append(fieldName).append("</b>") ;
		if(field!=null) bdr.append("<br/>").append(Annotations.getDependencyInjectionDescription(field)) ;
		bdr.append("</html>") ;
		
		return bdr.toString() ;
	}
	
	final static public String getMethodDocumentation(Class<?> clas, Signature stur)
	{
		// TODO: recuperer la Method dans BeanDef ou Signature (cf Bob)
		Method meth = null ;
		try
		{ meth = ReflectUtil.getMethod(clas, stur.getMethod()) ; }
		catch(Exception ex)
		{ /*ex.printStackTrace() ;*/ }
		
		StringBuilder bdr = new StringBuilder() ;
		bdr.append("<html>") ;
		bdr.append("<b>").append(stur).append("</b>") ;
		if(meth!=null)
		{
			bdr.append("<br/>") ;
			bdr.append(Annotations.getMethodDescription(meth)) ;
			
			for(Parameter para : Annotations.getParameters(meth))
				bdr.append("<br/>&nbsp;&nbsp;-&nbsp;").append(para) ;
			
			Class<?> retu = meth.getReturnType() ;
			if(retu!=void.class)
				bdr.append("<br/>").append("Returns ").append(retu.getSimpleName()).append(' ').append(Annotations.getMethodReturnType(meth)) ;	
		}
		
		bdr.append("</html>") ;
		
		return bdr.toString() ;
	}
    
	final static public String getClassName(Class<?> clas)
	{
		if(!clas.isArray()) return clas.getName() ;
		
		StringBuilder bdr =  new StringBuilder() ;
		
		Class<?> ctyp = clas.getComponentType() ;
		while(ctyp.isArray()) ctyp = ctyp.getComponentType() ;
		
		Package pakg = ctyp.getPackage() ;
		if(pakg!=null) bdr.append(pakg.getName()).append('.') ;
		
		bdr.append(clas.getSimpleName()) ;
		
		return bdr.toString() ;
	}
	
    final static public Class<?> getBoxedClass(Class<?> clas)
    {
    	for(Class<?> box : BOXED_CLASSES)
    	{
    		try
    		{
    			Field fild = box.getField("TYPE") ;
    			Class<?> type = (Class<?>)fild.get(null) ;
        		if(type.equals(clas)) return box ;
        	}
    		catch(Exception ex){}
       	}
    	
    	return null ;
    }
    
	final static public Method bestMatch(Class<?> c, String meth, Object[] args)
	{
		// 1 - Try to found a method that match with same parameter type
		Class[] types = NOARGS ;
		
		if(args!=null)
		{
			types = new Class[args.length] ;
			for(int i=0; i<types.length; i++)
				types[i] = args.getClass() ;
		}
		
		try
		{
			return c.getMethod(meth, types) ;
		}
		catch(NoSuchMethodException nsme)
		{
			// Next... 
		}
		
		// 2 - Try to find a method with the same name & same parameter number
		Method[] ms = c.getMethods() ;
		
		for(Method m : ms)
			if(meth.equals(m.getName())) 
				if(m.getParameterTypes().length == types.length)
					return m ;
		
		// 3 - Try to find a method with the same name only !
		for(Method m : ms)
			if(meth.equals(m.getName()))  return m ;

		// 4 - Not Found
		return null ;
	}

	final static public boolean hasParamTypes(Method m, Class<?>[] types)
	{
		if(types==null) types = new Class[0] ;
		                                  
		Class<?>[] cs = m.getParameterTypes() ;
		if(cs.length!=types.length) return false ;
		for(int i=0; i<types.length; i++)
		{
			if(!(types[i].isAssignableFrom(cs[i])))
				return false ;
		}
		return true ;
	}
		
	final static public Method[] getMethodWithParamTypes(Class c, Class<?>[] types)
	{
		Method[] all = c.getMethods() ;
		ArrayList<Method> alm = new ArrayList<Method>() ;
		for(Method m : all)
		{
			if(hasParamTypes(m, types))
				alm.add(m) ;
		}
		
		Method[] res = new Method[alm.size()] ;
		alm.toArray(res) ;
		return res ;
	}
    
    /* NO MORE : Regular expression
	final static public Method[] getSelectedMethods(Method[] meths, Collection<String> rgxs)
	throws PatternSyntaxException
	{
		if(meths==null) return null ;
		if(rgxs==null)  return null ; 
		
		HashMap<Method, Boolean> sel = new HashMap<Method, Boolean>() ;

		for(String regex : rgxs)
		{
			Pattern pat = Pattern.compile(regex);
			
			for(Method meth : meths)
			{
				if(match(meth, pat))
					sel.put(meth, Boolean.TRUE) ;
			}
		}

		Set<Method> sms = sel.keySet() ;
		Method[] methods = new Method[sms.size()] ;
		sms.toArray(methods) ;
		return methods ;
	}
	*/
}