/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.util ;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.*;

import javax.xml.transform.*;

import org.jdom.*;
import org.jdom.input.SAXBuilder;
import org.jdom.output.*;
import org.jdom.transform.*;
import org.jdom.xpath.XPath;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

public class JDOMUtil
{
	final static public SAXBuilder saxBuilder = new SAXBuilder() ;
	final static public XMLOutputter xmlout = new XMLOutputter(Format.getPrettyFormat()) ;
	
	static
	{
		saxBuilder.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false) ;
	}
	
	final static private Logger logr = Logger.getLogger("org.azuki.common") ;
	
	@SuppressWarnings("unchecked")
	final static public List<Element> getChildren(Element parent, boolean clon)
	{
		List<Element> chdn = parent.getChildren() ;
		if(!clon) return chdn ;
		
		List<Element> list = new ArrayList<Element>() ;
		for(Element lmen : chdn)
			list.add((Element)lmen.clone()) ;
		
		return list ;
	}
	
	final static public Element getElement(Element parent, String childName)
	{
		if(parent==null) return new Element(childName) ;
		
		Element child = parent.getChild(childName) ;
		if(child==null)
		{	child = new Element(childName) ;
			parent.addContent(child) ; }
		
		return child ;
	}
	
	final static public Attribute getAttribute(Element lmen, String attrName)
	{
		return getAttribute(lmen, attrName, "") ;
	}
	
	final static public Attribute getAttribute(Element lmen, String attrName, String defaultValue)
	{
		Attribute attr = lmen.getAttribute(attrName) ;
		if(attr==null)
		{	attr = new Attribute(attrName, defaultValue) ;
			lmen.setAttribute(attr) ; }
		
		return attr ;
	}
	
	final static public Element findChildWithAttribute(Element parent, String childName, String attrName, String attrValue)
	{
		return findChildWithAttribute(parent, childName, attrName, attrValue, true) ;
	}
	
	@SuppressWarnings("unchecked")
	final static public Element findChildWithAttribute(Element parent, String childName, String attrName, String attrValue, boolean clon)
	{
		Collection<Element> lmens = parent.getChildren(childName) ;
		return findElementWithAttribute(lmens, attrName, attrValue, clon) ;
	}
	
	final static public Element findElementWithAttribute(Collection<Element> lmens, String attrName, String attrValue)
	{
		return findElementWithAttribute(lmens, attrName, attrValue, true) ;
	}
	
	final static public Element findElementWithAttribute(Collection<Element> lmens, String attrName, String attrValue, boolean clon)
	{
		for(Element lmen : lmens)
		{
			String valu = lmen.getAttributeValue(attrName) ;
			if(attrValue.equals(valu))
				return clon ? (Element)lmen.clone() : lmen ;
		}
		return null ;
	}
	
	@SuppressWarnings("unchecked")
	final static public Map<String,String> mapAttributes(Element lmen)
	{
		Map<String,String> map = new HashMap<String,String>() ;
		
		List<Attribute> atts = lmen.getAttributes() ;
		for(Attribute attr : atts)
			map.put(attr.getName(), attr.getValue()) ;
		
		return map ;
	}
	
	final static public Map<String,String> mapAttributes(Element lmen, String... atts)
	{
		return mapAttributes(lmen, Arrays.asList(atts)) ;
	}
	
	final static public Map<String,String> mapAttributes(Element lmen, Collection<String> atts)
	{
		Map<String,String> map = new HashMap<String,String>() ;
		
		for(String attr : atts)
			map.put(attr, lmen.getAttributeValue(attr)) ;
		
		return map ;
	}
	
	final static public List<Element> listChildren(Element parent, String childName)
	{
		return listChildren(parent, childName, true) ;
	}
	
	@SuppressWarnings("unchecked")
	final static public List<Element> listChildren(Element parent, String childName, boolean clon)
	{
		List<Element> list = new ArrayList<Element>() ;
		
		List<Element> children = parent.getChildren(childName) ;
		for(Element child : children)
			list.add(clon ? (Element)child.clone() : child) ; 
		
		return list ;
	}
	
	final static public List<Element> listChildrenWithAttributes(Element parent, String childName, Map<String,String> atts)
	{
		return listChildrenWithAttributes(parent, childName, atts, true) ;
	}
	
	@SuppressWarnings("unchecked")
	final static public List<Element> listChildrenWithAttributes(Element parent, String childName, Map<String,String> atts, boolean clon)
	{
		List<Element> list = new ArrayList<Element>() ;
		
		List<Element> children = parent.getChildren(childName) ;
		for(Element child : children)
		{
			boolean eq = false ;
			
			for(Entry<String,String> ety : atts.entrySet())
			{	eq = ety.getValue().equals(child.getAttributeValue(ety.getKey())) ;
				if(!eq) break ; }
			
			if(eq) list.add(clon ? (Element)child.clone() : child) ;
		}
		
		return list ;
	}
	
	@SuppressWarnings("unchecked")
	final static public List<String> listChildrenAttributes(Element parent, String childName, String attrName)
	{
		List<String> list = new ArrayList<String>() ;
		
		List<Element> children = parent.getChildren(childName) ;
		for(Element child : children)
			list.add(child.getAttributeValue(attrName)) ; 
		
		return list ;
	}
	
	@SuppressWarnings("unchecked")
    final static public List<String> listAttributeNames(Element lmen)
	{
		List<String> list = new ArrayList<String>() ;
		
		List<Attribute> attrs = lmen.getAttributes() ;
		for(Attribute attr : attrs)
			list.add(attr.getName()) ;
		
		return list ;
	}
	
	final static public List<String> listAttributeValues(Collection<Element> lmens, String attrName, String defValue)
	{
		ArrayList<String> list = new ArrayList<String>() ;
		
		for(Element lmen : lmens)
		{
			String valu = lmen.getAttributeValue(attrName, defValue) ;
			if(!list.contains(valu))
				list.add(valu) ;
		}
		
		return list ;
	}
	
	/*
	 * XSL transformation
	 */
//	final static public List xslTransform(Document xmlSource, Document stylesheet)
//	throws XSLTransformException
//	{
//		XSLTransformer xslt = new XSLTransformer(stylesheet) ;
//		return xslt.transform(xmlSource) ;
//	}
	
	static private TransformerFactory tf = null ; 
	
	final static public JDOMResult xslTransform(Document xmlSource, Document stylesheet)
	throws Exception
	{		
		if(tf==null) tf = TransformerFactory.newInstance() ;
		
		Transformer t = tf.newTransformer(new JDOMSource(stylesheet)) ;
		
		JDOMSource jsrc = new JDOMSource(xmlSource) ;
		JDOMResult jres = new JDOMResult() ;
		
		t.transform(jsrc, jres) ;
		
		return jres ;
	}
	
	/*
	 * I/O
	 */
	final static public Document read(ClassLoader cl, String xmlResource)
	throws JDOMException, IOException
	{
		InputStream is = cl.getResourceAsStream(xmlResource) ;
		if(is==null) logr.severe("Document not found ("+xmlResource+")") ;
		
		return is==null ? null : read(is) ;
	}
	
	final static public Document read(File file)
	throws JDOMException, IOException
	{
		return saxBuilder.build(file) ;
	}
	
	final static public Document read(InputStream xmlInput)
	throws JDOMException, IOException
	{
		return saxBuilder.build(xmlInput) ;
	}
	
	final static public Document parseDocument(String s)
	throws JDOMException, IOException
	{
		StringReader sr = new StringReader(s) ;
		Document doc = saxBuilder.build(sr) ;
		sr.close() ;
		return doc ;
	}
	
	final static public Element parseElement(String s)
	throws JDOMException, IOException
	{
		return parseDocument(s).detachRootElement() ;
	}
	
	final static public void write(Document doc, File file)
	{
		String filePath = file.getPath() ;
		
		logr.finest("Writting ("+filePath+")") ;
		try
		{
			if(file.exists())
			{
				File bkup = new File(filePath+".bak") ;
				if(bkup.exists()) bkup.delete() ;
				file.renameTo(bkup) ;
			}
			
			file = new File(filePath) ;
			if(!file.getParentFile().exists())
				file.getParentFile().mkdirs() ;
			
			FileOutputStream fos = new FileOutputStream(file) ;
			
			xmlout.output(doc, fos) ;
			
			fos.close() ;
		}
		catch(Exception ex)
		{
			logr.log(Level.SEVERE, "Error while writting ("+filePath+")", ex) ;
		}
	}
	
	/*
	 * XPath
	 */
	@SuppressWarnings("unchecked")
	final static public List<Element> query(Element lmen, String xpathQuery)
	throws JDOMException
	{
		Document doc = null ;
		if(lmen.getDocument()==null)
			doc = new Document().setRootElement(lmen) ;
		
		List<Element> list = XPath.selectNodes(lmen, xpathQuery) ;
		
		if(doc!=null)
		{	lmen.detach() ;
			doc = null ; }
		
		return list ;
	}
	
	@SuppressWarnings("unchecked")
	final static public List<Element> query(Document doc, String xpathQuery)
	throws JDOMException
	{
		return XPath.selectNodes(doc, xpathQuery) ;
	}
}