/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.widget ;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.tree.*;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

@SuppressWarnings("serial")
public class JCheckTree extends JTree implements ItemSelectable
{
	final protected MouseListener mouseSelectionAdapter = new MouseSelectionAdapter() ;
	
	final private Set<TreePath> checkedPaths = new TreeSet<TreePath>(new TreePathComparator()) ;
	
	private boolean checkable = true ; 
	private TreeCellRenderer treeCellRenderer = new DefaultTreeCellRenderer() ;
	
	public JCheckTree(Object... value)
	{
		this(JTree.createTreeModel(value)) ;
	}
	
	public JCheckTree(Hashtable<?,?> value)
	{
		this(JTree.createTreeModel(value)) ;
	}
	
	public JCheckTree(Vector<?> value)
	{
		this(JTree.createTreeModel(value)) ;
	}
	
	public JCheckTree(TreeNode root)
	{
		this(root, false) ;
	}
	
	public JCheckTree(TreeNode root, boolean askAllowsChildren)
	{
		this(new DefaultTreeModel(root, askAllowsChildren)) ;
	}
	
	public JCheckTree(TreeModel model)
	{
		super(model) ;
		
		addMouseListener(mouseSelectionAdapter) ;
	}
	
	final public boolean isCheckable()
	{
		return checkable ;
	}
	
	final public void setCheckable(boolean checkable)
	{
		if(this.checkable==checkable) return ;
		
		checkedPaths.clear() ;
		
		boolean oldCheckable = this.checkable ;
		
		this.checkable = checkable ;
		firePropertyChange("checkable", oldCheckable, checkable) ;
		invalidate() ;
		
		setCellRenderer(treeCellRenderer) ;
	}
	
	final public boolean isCheckedEmpty()
	{
		return checkedPaths.isEmpty() ;
	}
	
	final public Collection<TreePath> getCheckedPaths()
	{
		return Collections.unmodifiableCollection(checkedPaths) ;
	}
	
	final public void setCheckedPath(TreePath path, boolean checked)
	{
		if(!checkable) return ;
		
		if(path==null) return ;
		
		// current node
		if(checked) checkPath(path) ;
		else		uncheckPath(path) ;
		
		TreeNode node = (TreeNode)path.getLastPathComponent() ;
		for(int i=0 ; i<node.getChildCount() ; i++)
			setCheckedPath(path.pathByAddingChild(node.getChildAt(i)), checked) ;
		
		repaint() ;
		
		// parent node
		TreePath parent = path.getParentPath() ;
		if(parent==null) return ;
		if(checked==checkedPaths.contains(parent)) return ;
		
		while(parent!=null)
		{
			SwingToolkit.expandPath(this, parent) ;
			
			if(!checked)						uncheckPath(parent) ;
			else if(isChildrenChecked(parent)) 	checkPath(parent) ;

			parent = parent.getParentPath() ;
		}
		
		repaint() ;
	}
	
	final private boolean isChildrenChecked(TreePath parent)
	{
		if(parent==null) return false ;
		
		TreeNode node = (TreeNode)parent.getLastPathComponent() ;
		for(int i=0 ; i<node.getChildCount() ; i++)
		{
			TreePath path = parent.pathByAddingChild(node.getChildAt(i)) ;
			if(!checkedPaths.contains(path)) return false ;
		}
		
		return true ;
	}
	
	final private void checkPath(TreePath path)
	{
		if(checkedPaths.add(path))
			notifyCheckChange(path, true) ;
	}
	
	final private void uncheckPath(TreePath path)
	{
		if(checkedPaths.remove(path))
			notifyCheckChange(path, false) ;
	}
	
	final protected void notifyCheckChange(TreePath path, boolean checked)
	{
		int status = checked ? ItemEvent.SELECTED : ItemEvent.DESELECTED ;
		ItemEvent ie = new ItemEvent(this, ItemEvent.ITEM_STATE_CHANGED, path, status) ;
		
		ItemListener[] lsnrs = listenerList.getListeners(ItemListener.class) ;
		for(ItemListener lsnr : lsnrs)
			lsnr.itemStateChanged(ie) ;
	}
	
	@Override
	final public void setCellRenderer(TreeCellRenderer treeCellRenderer)
	{
		this.treeCellRenderer = treeCellRenderer ;
		
		TreeCellRenderer rndr = checkable ? new CheckTreeCellRenderer(treeCellRenderer) : treeCellRenderer ;
		super.setCellRenderer(rndr) ;
	}
	
	/*
	 * ItemSelectable
	 */
	final public Object[] getSelectedObjects()
	{
		return checkedPaths.toArray() ;
	}
	
	final public void addItemListener(ItemListener l)
	{
		listenerList.add(ItemListener.class, l) ;
	}

	final public void removeItemListener(ItemListener l)
	{
		listenerList.remove(ItemListener.class, l) ;
	}
	
	/**
	 * Mouse selection listener
	 */
	final class MouseSelectionAdapter implements MouseListener
	{
		/*
		 * MouseListener
		 */
		final public void mouseClicked(MouseEvent e)
		{
			if(e.getClickCount() > 1) return ; //FIXME: Timer ??
			
			TreePath path = getPathForLocation(e.getX(), e.getY()) ;
			setCheckedPath(path, !checkedPaths.contains(path)) ;
		}
		final public void mouseEntered(MouseEvent e) {}
		final public void mouseExited(MouseEvent e) {}
		final public void mousePressed(MouseEvent e) {}
		final public void mouseReleased(MouseEvent e) {}
	} 
	
	/**
	 * Tree cell selection renderer
	 */
	final class CheckTreeCellRenderer extends JComponent implements TreeCellRenderer
	{
		final private TreeCellRenderer treeRenderer ;
		final private JCheckBox checkBox ;
		
		public CheckTreeCellRenderer(TreeCellRenderer treeRenderer)
		{
			this.treeRenderer = treeRenderer ;
			
			checkBox = new JCheckBox() ;
			checkBox.setOpaque(false) ;
			checkBox.setMargin(new Insets(0,0,0,0)) ;
			
			setLayout(new BorderLayout()) ;
			add(checkBox, BorderLayout.WEST) ;
			add(Box.createGlue(), BorderLayout.EAST) ;
		}
		
		final public TreeCellRenderer getCellRenderer()
		{
			return treeRenderer ;
		}
		
		/*
		 * TreeCellRenderer
		 */
		final public Component getTreeCellRendererComponent(JTree tree, Object value, boolean selected, boolean expanded, boolean leaf, int row, boolean hasFocus)
		{	
			Collection<TreePath> paths = ((JCheckTree)tree).getCheckedPaths() ;
			
			TreePath path = tree.getPathForRow(row) ;
			
			// CheckBox
			checkBox.setSelected(paths!=null && path!=null && paths.contains(path)) ;
			
			// Renderer
			Component comp = treeRenderer.getTreeCellRendererComponent(tree, value, selected, expanded, leaf, row, hasFocus) ;
			remove(1) ;
			add(comp) ;
			
			setComponentOrientation(tree.getComponentOrientation()) ;

			return this ;
		}
	}
	
	/**
	 * Tree path comparator
	 */
	final public class TreePathComparator implements Comparator<TreePath>
	{
		/*
		 * Comparator
		 */
		final public int compare(TreePath tp1, TreePath tp2)
		{
			if(tp1==null && tp2==null) return 0 ;
			return String.valueOf(tp1).compareTo(String.valueOf(tp2)) ;
		}
	}
}