/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.widget ;

import java.awt.*;
import java.util.Enumeration;

import javax.swing.*;
import javax.swing.table.TableCellEditor;
import javax.swing.tree.*;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

public class SwingToolkit
{
	final static public <T extends Component> T findAncestor(Component c, Class<T> clas)
	{
		if(clas.isInstance(c)) return clas.cast(c) ;
		
		c = c.getParent() ;
		return c==null ? null : findAncestor(c, clas) ;
	}
	
	final static public void ensureEditingStopped(Container cont)
	{
		for(Component c : cont.getComponents())
		{
			if(c instanceof JTable)
			{
				TableCellEditor tce = ((JTable)c).getCellEditor() ;
				if(tce!=null) tce.stopCellEditing() ;
				continue ;
			}
			else if(c instanceof JTree)
			{
				TreeCellEditor tce = ((JTree)c).getCellEditor() ;
				if(tce!=null) tce.stopCellEditing() ;
				continue ;
			}
			else if(c instanceof Container)
				ensureEditingStopped((Container)c) ;
		}
	}
	
	/*
	 * Tree
	 */
	final static public void expandAll(JTree tree)
	{
		Object root = tree.getModel().getRoot() ;
		expandAll(tree, new TreePath(root)) ;
	}
	
	final static public void expandAll(JTree tree, TreePath path)
	{
		setDescendantExpandedState(tree, path, true) ;
	}
	
	final static public void collapseAll(JTree tree)
	{
		Object root = tree.getModel().getRoot() ;
		collapseAll(tree, new TreePath(root)) ;
	}
	
	final static public void collapseAll(JTree tree, TreePath path)
	{
		setDescendantExpandedState(tree, path, false) ;
	}
	
	final static public void setDescendantExpandedState(JTree tree, TreePath path, boolean state)
	{
		int row = tree.getRowForPath(path) ;
		
		// otherwise only first-level are expanded
		if(state) tree.expandRow(row) ;
		
		TreeNode node = (TreeNode)path.getLastPathComponent() ;
		for(int i=0 ; i<node.getChildCount() ; i++)
		{
			TreePath temp = path.pathByAddingChild(node.getChildAt(i)) ;
			setDescendantExpandedState(tree, temp, state) ;
		}
		
		// otherwise only first-level are collapsed
		if(!state && row!=0) tree.collapseRow(row) ;
	}
	
	final static public void expandPath(JTree tree, TreePath path)
	{
		TreePath temp = null ;
		
		Object[] objs = path.getPath() ;
		for(Object obj : objs)
		{
			if(temp==null) 	temp = new TreePath(obj) ;
			else			temp = temp.pathByAddingChild(obj) ;

			int row = tree.getRowForPath(temp) ;
			if(row==-1) temp = findTreePath(tree, temp) ;
			
			row = tree.getRowForPath(temp) ;
			if(row==-1) break ;

			tree.expandRow(row) ;
		}
	}
	
	final static public TreePath findTreePath(JTree tree, TreePath path)
	{
		TreeNode node = (TreeNode)tree.getModel().getRoot() ;
		
		boolean first = true ;
		
		for(Object obj : path.getPath())
		{
			if(first) continue ;
			
			first = false ;
			
			Enumeration enu = node.children() ;
			while(true)
			{
				if(!enu.hasMoreElements())
					return null ; // not found ! 
				
				Object chld = enu.nextElement() ;
				if(obj.equals(chld))
				{	node = (TreeNode)chld ;
					break ; }
			}
		}
		
		return getTreePath(node) ;
	}
	
	final static public TreePath getTreePath(TreeNode node)
	{
		TreeNode parent = node.getParent() ;
		if(parent==null) return new TreePath(node) ;
		
		TreePath ppat = getTreePath(parent) ;
		return ppat.pathByAddingChild(node) ;
	}
}