/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.common.widget.model ;

import java.util.*;

import javax.swing.event.*;
import javax.swing.table.TableModel;

/**
 * @author Yoann Perrin (yoann.perrin@gmail.com)
 */

public class ItemChooserTableModel<T> implements TableModel
{
	public enum ItemSelection {ONLY_ONE, AT_LEAST_ONE, UNCHECKED}
	
	final protected EventListenerList modelListenerList = new EventListenerList() ;
	
	final private String label ;
	final private List<T> itemList ;
	final private List<Boolean> booleans ;
	
	private ItemSelection itemSelection = ItemSelection.UNCHECKED ;
	
	public ItemChooserTableModel(String label, List<T> itemList)
	{
		this(label, itemList, Boolean.TRUE) ;
	}
	
	public ItemChooserTableModel(String label, List<T> itemList, Boolean defaultSelection)
	{
		this.label = label ;
		this.itemList = itemList ;
		
		booleans = new ArrayList<Boolean>(itemList.size()) ;
		for(int i=0 ; i<itemList.size() ; i++)
			booleans.add(defaultSelection) ;
	}

	final public ItemSelection getSelectionMode()
	{
		return itemSelection ;
	}
	
	final public void setSelectionMode(ItemSelection itemSelection)
	{
		this.itemSelection = itemSelection ;
	}
	
	final public void selectAll()
	{
		Collections.fill(booleans, Boolean.TRUE) ;
		fireModelChanged(0, getRowCount()-1) ;
	}
	
	final public void clearSelection()
	{
		Collections.fill(booleans, Boolean.FALSE) ;
		fireModelChanged(0, getRowCount()-1) ;
	}
	
	final public boolean isSelected(T item)
	{
		int index = itemList.indexOf(item) ;
		return isSelected(index) ;
	}
	
	final public boolean isSelected(int index)
	{
		return index>=0 && index<booleans.size() && booleans.get(index).booleanValue() ;
	} 
	
	final public void setSelected(Collection<T> c, Boolean b)
	{
		for(T t : c)
		{	int idx = itemList.indexOf(t) ;
			if(idx!=-1) booleans.set(idx, b) ; }
		
		fireModelChanged(0, getRowCount()-1) ;
	}
	
	final public boolean isSelectionValid()
	{
		int count = getSelectedItems().size() ;
		
		if(itemSelection==ItemSelection.ONLY_ONE) return count==1 ;
		else if(itemSelection==ItemSelection.AT_LEAST_ONE) return count>=1 ;
		
		return true ;
	}
	
	final public T getItem(int index)
	{
		return index>=0 && index<itemList.size() ? itemList.get(index) : null ;
	}
	
	final public List<T> getItems()
	{
		return new ArrayList<T>(itemList) ;
	}
	
	final public List<T> getSelectedItems()
	{
		List<T> list = new ArrayList<T>() ;
		
		for(int i=0 ; i<booleans.size() ; i++)
		{
			if(booleans.get(i).booleanValue())
				list.add(itemList.get(i)) ;
		}
		
		return list ;
	}
	
	final protected void fireModelChanged(int row)
	{
		TableModelEvent tme = new TableModelEvent(this, row) ;
		
		TableModelListener[] listeners = modelListenerList.getListeners(TableModelListener.class) ; 
		for(TableModelListener tml : listeners)
			tml.tableChanged(tme) ;
	}
	
	final protected void fireModelChanged(int firstRow, int lastRow)
	{
		TableModelEvent tme = new TableModelEvent(this, firstRow, lastRow) ;
		
		TableModelListener[] listeners = modelListenerList.getListeners(TableModelListener.class) ; 
		for(TableModelListener tml : listeners)
			tml.tableChanged(tme) ;
	}
	
	
	/*
	 * TableModel
	 */
	final public Class<?> getColumnClass(int columnIndex)
	{
		return columnIndex==0 ?  Boolean.class : Object.class ; // FIXME
	}

	final public int getColumnCount()
	{
		return 2 ;
	}

	final public String getColumnName(int columnIndex)
	{
		return columnIndex==0 ? "" : label ;
	}

	final public int getRowCount()
	{
		return itemList.size() ;
	}

	final public Object getValueAt(int rowIndex, int columnIndex)
	{
		List l = columnIndex==0 ? booleans : itemList ;
		return l.get(rowIndex) ;
	}

	final public boolean isCellEditable(int rowIndex, int columnIndex)
	{
		return columnIndex==0 ;
	}

	final public void setValueAt(Object aValue, int rowIndex, int columnIndex)
	{
		if(columnIndex==1) return ;
		
		Boolean val = (Boolean)aValue ;
		
		if(itemSelection==ItemSelection.ONLY_ONE && val.booleanValue()) clearSelection() ;
		booleans.set(rowIndex, (Boolean)aValue) ;
		
		fireModelChanged(rowIndex) ;
	}
	
	final public void addTableModelListener(TableModelListener l)
	{
		modelListenerList.add(TableModelListener.class, l) ;
	}
	
	final public void removeTableModelListener(TableModelListener l)
	{
		modelListenerList.remove(TableModelListener.class, l) ;
	}
}