/*
 *   Created the 2 April 2006
 *
 *   Azuki Framework
 *   Copyright 2006 the original author or authors.
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *   
 *   Contact : support@azuki-framework.org
 */

package org.azuki.framework.context;

import java.io.*;
import java.util.logging.*;

import org.azuki.framework.Framework;

/**
 *   @author Robert Bakic (robert.bakic@gmail.com)
 */

final public class Log 
{
	final static private Logger initlog = Logger.getAnonymousLogger() ;
	
	static
	{
		initlog.setUseParentHandlers(false) ;
		initlog.addHandler(new AzukiLogHandler(System.out, Level.ALL)) ;
		try
		{ initlog.addHandler(new AzukiLogHandler(new FileOutputStream("initlog.log", false), Level.ALL)) ; }
		catch(Exception t)
		{ t.printStackTrace() ; }
	}
	
	final static public Logger getLogger()
	{
		if(!Framework.inited()) return initlog ;
		
		try
		{
			Logger lgr = (Logger)ThreadContext.lookup(IContext.OBJECT_LOGGER) ;
			if(lgr!=null) return lgr ; // Logger resource correctly set
		}
		catch(Exception e)
		{ 
			System.out.println("Cannot retrieve logger from context: "+e.getMessage()) ;
		}
		
		// reach only if no "LoggerFactory" are correctly configured !
		return null ;
	}	

	//=======================================================================
    // Start of simple convenience methods using level names as method names
    //=======================================================================

    /**
     * Log a SEVERE message.
     * <p>
     * If the logger is currently enabled for the SEVERE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    final static public void severe(String msg)
    {
    	logp(Level.SEVERE, msg, null) ;
    }
    
    /**
     * Log a SEVERE message.
     * <p>
     * If the logger is currently enabled for the SEVERE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   thr	The thrown exception
     */
    final static public void severe(Throwable thr)
    {
    	logp(Level.SEVERE, Log.getStack(thr), null) ;
    }

    /**
     * Log a WARNING message.
     * <p>
     * If the logger is currently enabled for the WARNING message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    final static public void warning(String msg)
    {
    	logp(Level.WARNING, msg, null) ;
    }

    /**
     * Log an INFO message.
     * <p>
     * If the logger is currently enabled for the INFO message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    final static public void info(String msg)
    {
    	logp(Level.INFO, msg, null) ;
    }

    /**
     * Log a CONFIG message.
     * <p>
     * If the logger is currently enabled for the CONFIG message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    final static public void config(String msg)
    {
    	logp(Level.CONFIG, msg, null) ;
    }

    /**
     * Log a FINE message.
     * <p>
     * If the logger is currently enabled for the FINE message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    final static public void fine(String msg)
    {
    	logp(Level.FINE, msg, null) ;
    }

    /**
     * Log a FINER message.
     * <p>
     * If the logger is currently enabled for the FINER message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    final static public void finer(String msg)
    {
    	logp(Level.FINER, msg, null) ;
    }

    /**
     * Log a FINEST message.
     * <p>
     * If the logger is currently enabled for the FINEST message 
     * level then the given message is forwarded to all the
     * registered output Handler objects.
     * <p>
     * @param   msg	The string message (or a key in the message catalog)
     */
    final static public void finest(String msg)
    {
    	logp(Level.FINEST, msg, null) ;
    }
    
    //================================================================
    // End of convenience methods 
    //================================================================
    final static private void logp(Level lvl, String msg, Throwable thr)
	{
    	Logger logr = getLogger() ;
    	if(logr==null || logr.getHandlers().length==0) return ;
    	
    	String clas = "" ;
    	String meth = "" ;
    	
    	if(lvl.intValue()>=Level.WARNING.intValue())
    	{
    		StackTraceElement lastCall = getLastCalled() ;
    		clas = lastCall.getClassName() ;
    		meth = lastCall.getMethodName() ;
    		if(msg==null && thr!=null) msg = thr.getMessage() ;
    	}
		
		logr.logp(lvl, clas, meth, msg, thr) ;
	}
    
	final static private String loggerClass = Logger.class.getName() ; //"java.util.logging.Logger"
    final static private String loggerResourceClass = Log.class.getName() ; //"org.azuki.framework.context.Log"
    final static private StackTraceElement getLastCalled()
	{
    	// Get the stack trace.
		StackTraceElement stack[] = new Throwable().getStackTrace() ;
    	
    	int ix = stack.length-1 ;
		while(ix>=0)
		{
		    StackTraceElement frame = stack[ix--] ;
		    String cname = frame.getClassName() ;
			
		    if(cname.equals(loggerClass)
		    || cname.equals(loggerResourceClass))
		    {
		        return stack[ix+2] ;
		    }
		}
		
		return null ;
	}
	
    final static public String getStack(Throwable t)
	{
		// Stockage dans une string de la pile d'éxecution
		String failure ;

		ByteArrayOutputStream baos = new ByteArrayOutputStream() ;

		try
		{
			PrintStream p = new PrintStream(baos);
			t.printStackTrace(p);
			failure = new String(baos.toByteArray()) ;
			p.close();
			baos.close();
		}
		catch(Exception e2)
		{
			failure = t.getMessage() ;
		}
		
		return failure ;
	}

    //================================================================
    // Default handler implementation
    //================================================================
    final static public class AzukiLogHandler extends Handler
    {
    	final private PrintWriter pw ;
    	
    	public AzukiLogHandler(OutputStream os, Level levl)
    	{
    		this(new PrintWriter(os), levl) ;
    	}
    	
    	public AzukiLogHandler(PrintWriter pw, Level levl)
    	{
    		this.pw = pw ;
    		
    		setLevel(levl) ;
    	}
    	
    	@Override
    	final public void publish(LogRecord rec)
    	{
    		if(!isLoggable(rec)) return ;
    		
    		if(rec.getLevel()==Level.WARNING) 		pw.print("WNG: ") ;
    		else if(rec.getLevel()==Level.SEVERE) 	pw.print("SVR: ") ;
    		
    		pw.println(rec.getMessage()) ;
    		pw.flush() ;
    	}

    	@Override
    	final public void flush()
    	{
    		pw.flush() ;
    	}

    	@Override
    	final public void close()
    	throws SecurityException
    	{
    		pw.close() ;
    	}
    }
}